"""Contains the code responsible for representing the structure of a model as a graph."""

from collections.abc import Callable
from enum import Enum
from typing import Any, Self
from networkx import DiGraph

from stormvogel.model import Action, EmptyAction, Model, State, Value

ACTION_ID_OFFSET = 10**10  # Offset to avoid ID clashes between state and action nodes


class NodeType(Enum):
    STATE = 0
    ACTION = 1
    UNDEFINED = 2


class ModelGraph(DiGraph):
    """A directed graph describing the structure of a model.

    States and actions (except EmptyActions) are represented as nodes in the graph.
    All outgoing edges of a state node describe the available actions for that state.
    The outgoing edges from an action describe the possible next states (possible choices)
    and hold the probability of each transition as a node attribute.
    """

    def __init__(
        self,
    ) -> None:
        super().__init__(self)
        self._state_action_id_map: dict[tuple[int, Action], int] = dict()
        self._next_action_id = ACTION_ID_OFFSET

    @property
    def state_action_id_map(self) -> dict[tuple[int, Action], int]:
        """
        dict[tuple[int, Action], int]: A mapping from (state ID, Action) pairs to internal action node IDs.

        This dictionary allows lookup of the internal graph node ID corresponding to a given (state, action) pair.
        It is used to associate actions with their respective nodes in the graph structure.
        """
        return self._state_action_id_map

    def add_state(self, state: State | int, **attr):
        """
        Adds a state node to the graph.

        If a `State` object is provided, its `id` is used as the node identifier.
        Additional attributes can be assigned to the state node via keyword arguments.

        Args:
            state (State | int): The state to add, either as a `State` object or its ID.
            **attr: Arbitrary keyword arguments representing attributes to associate with the state node.
        """
        if isinstance(state, State):
            state = state.id
        self.add_node(state, type=NodeType.STATE, **attr)

    def add_action(self, state: State | int, action: Action, **action_attr):
        """
        Adds an action node to the graph and connects it to a given state.

        The action node is uniquely identified and linked from the source state.
        The action is skipped if it is an `EmptyAction`. Additional attributes
        can be assigned to the action node via keyword arguments.

        Args:
            state (State | int): The source state (either a `State` object or its ID) from which the action originates.
            action (Action): The action to add.
            **action_attr: Arbitrary keyword arguments representing attributes to associate with the action node.

        Raises:
            AssertionError: If the source state is not already in the graph.
        """
        if isinstance(state, State):
            state = state.id
        assert state in self.nodes, f"State {state} not in graph yet"
        if action == EmptyAction:
            return
        self.add_node(self._next_action_id, type=NodeType.ACTION, **action_attr)
        self.add_edge(state, self._next_action_id)
        self._state_action_id_map[(state, action)] = self._next_action_id
        self._next_action_id += 1

    def add_transition(
        self,
        state: State | int,
        action: Action,
        next_state: int | State,
        probability: Value,
        **attr,
    ) -> None:
        """
        Adds a transition to the graph with an associated probability.

        For non-empty actions, this adds an edge from the action node to the target state.
        For `EmptyAction`, the edge is added directly from the source state to the target state.
        Additional attributes can be provided for the transition edge.

        Args:
            state (State | int): The source state (either a `State` object or its ID).
            action (Action): The action that causes the transition.
            next_state (int | State): The target state reached by the transition.
            probability (Value): The probability associated with the transition.
            **attr: Arbitrary keyword arguments representing attributes to associate with the transition edge.

        Raises:
            AssertionError: If the source state or target state is not in the graph,
                or if the action node is missing (for non-empty actions).
        """
        if isinstance(state, State):
            state = state.id
        if isinstance(next_state, State):
            next_state = next_state.id

        action_id = self._state_action_id_map.get((state, action), None)
        assert state in self.nodes, "State {state} not in graph."
        assert (
            action == EmptyAction or action_id is not None
        ), "Action node for action {action} in state {state} not in graph."
        assert next_state in self.nodes, "Next state {next_state} not in graph."

        if action == EmptyAction:
            self.add_edge(state, next_state, probability=probability, **attr)
        else:
            self.add_edge(action_id, next_state, probability=probability, **attr)

    @classmethod
    def from_model(
        cls,
        model: Model,
        state_properties: Callable[[State], dict[str, Any]] | None = None,
        action_properties: Callable[[State, Action], dict[str, Any]] | None = None,
        transition_properties: (
            Callable[[State, Action, State], dict[str, Any]] | None
        ) = None,
    ) -> Self:
        """
        Constructs a directed graph representation of a Markov Decision Process (MDP) from a model instance.

        This method initializes the graph from the provided `model` by adding all states, actions, and choices.
        Optional callbacks allow customization of properties for states, actions, and choices.

        Args:
            model (Model): The MDP model containing states and choices.
            state_properties (Callable[[State], dict[str, Any]], optional): A callable that returns a dictionary
                of properties for a given state. Defaults to None.
            action_properties (Callable[[State, Action], dict[str, Any]], optional): A callable that returns a
                dictionary of properties for a given action from a state. Defaults to None.
            transition_properties (Callable[[State, Action, State], dict[str, Any]], optional): A callable that
                returns a dictionary of properties for a transition from a source state via an action to a target
                state. Defaults to None.

        Returns:
            Self: An instance of the graph populated with the states, actions, and choices from the model.

        Examples:
            >>> import stormvogel.examples as examples
            >>> mdp = examples.create_lion_mdp()
            >>> G = ModelGraph.from_model(mdp, state_properties = lambda s: {"labels": s.labels})
            >>> G.nodes[mdp.get_initial_state().id]
            {'type': <NodeType.STATE: 0>, 'labels': ['init']}
        """
        G = cls()
        for _, state in model:
            props = dict()
            if state_properties is not None:
                props = state_properties(state)
            G.add_state(state, **props)

        for state_id, choice in model.choices.items():
            state = model.get_state_by_id(state_id)
            for action, branch in choice:
                action_props = dict()
                if action_properties is not None:
                    action_props = action_properties(state, action)
                G.add_action(state_id, action, **action_props)
                for probability, target in branch:
                    transition_props = dict()
                    if transition_properties is not None:
                        transition_props = transition_properties(state, action, target)
                    G.add_transition(
                        state_id,
                        action,
                        target,
                        probability=probability,
                        **transition_props,
                    )
        return G



"""Simple implementations of two model checking algorithms in stormvogel,
along with a function to display the workings of the algorithms."""

from typing import Any
import stormvogel.model
from time import sleep


def naive_value_iteration(
    model: stormvogel.model.Model, epsilon: float, target_state: stormvogel.model.State
) -> list[list[stormvogel.model.Value]]:
    """Run naive value iteration. The result is a 2D list where result[n][m] is the probability to be in state m at step n.

    Args:
        model (stormvogel.model.Model): Target model.
        steps (int): Amount of steps.
        target_state (stormvogel.model.State): Target state of the model.

    Returns:
        list[list[float]]: The result is a 2D list where result[n][m] is the value of state m at iteration n.
    """
    if epsilon <= 0:
        RuntimeError("The algorithm will not terminate if epsilon is zero.")

    # Create a dynamic matrix (list of lists) to store the result.
    values_matrix = [[0 for state in model.get_states()]]
    values_matrix[0][target_state.id] = 1

    terminate = False
    while not terminate:
        old_values = values_matrix[len(values_matrix) - 1]
        new_values = [None for state in model.get_states()]
        for sid, state in model:
            choices = model.get_choice(state)
            # Now we have to take a decision for an action.
            action_values = {}
            for action, branch in choices:
                branch_value = sum(
                    [prob * old_values[state.id] for (prob, state) in branch]  # type: ignore
                )
                action_values[action] = branch_value
            # We take the action with the highest value.
            new_values[sid] = max(action_values.values())
        values_matrix.append(new_values)  # type: ignore
        terminate = (
            sum([abs(x - y) for (x, y) in zip(new_values, old_values)]) < epsilon  # type: ignore
        )
    return values_matrix  # type: ignore


def dtmc_evolution(model: stormvogel.model.Model, steps: int) -> list[list[float]]:
    """Run DTMC evolution. The result is a 2D list where result[n][m] is the probability to be in state m at step n.

    Args:
        model (stormvogel.model.Model): Target model.
        steps (int): Amount of steps.

    Returns:
        list[list[float]]: The result is a 2D list where result[n][m] is the probability to be in state m at step n.
    """
    if steps < 2:
        RuntimeError("Need at least two steps")
    if model.type != stormvogel.model.ModelType.DTMC:
        RuntimeError("Only works for DTMC")

    # Create a matrix and set the value for the starting state to 1 on the first step.
    matrix_steps_states = [[0.0 for s in model.get_states()] for x in range(steps)]
    matrix_steps_states[0][model.get_initial_state().id] = 1

    # Apply the updated values for each step.
    for current_step in range(steps - 1):
        next_step = current_step + 1
        for s_id, s in model:
            branch = model.get_branches(s)
            for transition_prob, target in branch:
                current_prob = matrix_steps_states[current_step][s_id]
                assert isinstance(transition_prob, (int, float))
                matrix_steps_states[next_step][target.id] += current_prob * float(
                    transition_prob
                )

    return matrix_steps_states


def invert_2d_list(li: list[list[Any]]) -> list[list[Any]]:
    res = []
    for i in range(len(li[0])):
        sublist = []
        for j in range(len(li)):
            sublist.append(li[j][i])
        res.append(sublist)
    return res


def display_value_iteration_result(
    res: list[list[float]], hor_size: int, labels: list[str]
):
    """Display a value iteration results using matplotlib.

    Args:
        res (list[list[float]]): 2D list where result[n][m] is the probability to be in state m at step n.
        hor_size (int): the horizontal size of the plot, in inches.
        labels (list[str]): the names of all the states.
    """
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots(1, 1)
    ax.set_xticks(range(len(res)))
    ax.set_yticks(range(len(res[0]) + 1))
    ax.set_yticklabels(labels + [""])

    ax.imshow(invert_2d_list(res), cmap="hot", interpolation="nearest", aspect="equal")
    plt.xlabel("steps")
    plt.ylabel("states")
    fig.set_size_inches(hor_size, hor_size)

    plt.show()


def arg_max(funcs, args):
    """Takes a list of callables and arguments and return the argument that yields the highest value."""
    executed = [f(x) for f, x in zip(funcs, args)]
    index = executed.index(max(executed))
    return args[index]


def policy_iteration(
    model: stormvogel.model.Model,
    prop: str,
    visualize: bool = True,
    layout: stormvogel.layout.Layout = stormvogel.layout.DEFAULT(),
    delay: int = 2,
    clear: bool = True,
) -> stormvogel.Result:
    """Performs policy iteration on the given mdp.
    Args:
        model (Model): MDP.
        prop (str): PRISM property string to maximize. Rembember that this is a property on the induced DTMC, not the MDP.
        visualize (bool): Whether the intermediate and final results should be visualized. Defaults to True.
        layout (Layout): Layout to use to show the intermediate results.
        delay (int): Seconds to wait between each iteration.
        clear (bool): Whether to clear the visualization of each previous iteration.
    """
    old = None
    new = stormvogel.random_scheduler(model)

    while not old == new:
        old = new

        dtmc = old.generate_induced_dtmc()
        dtmc_result = stormvogel.model_checking(dtmc, prop=prop)  # type: ignore

        if visualize:
            vis = stormvogel.visualization.JSVisualization(
                model, layout=layout, scheduler=old, result=dtmc_result
            )
            vis.show()
            sleep(delay)
            if clear:
                vis.clear()

        choices = {
            i: arg_max(
                [
                    lambda a: sum(
                        [
                            (p * dtmc_result.get_result_of_state(s2.id))  # type: ignore
                            for p, s2 in s1.get_outgoing_choice(a)  # type: ignore
                        ]
                    )
                    for _ in s1.available_actions()
                ],
                s1.available_actions(),
            )
            for i, s1 in model
        }
        new = stormvogel.Scheduler(model, choices)
    if visualize:
        print("Value iteration done:")
        stormvogel.show(model, layout=layout, scheduler=new, result=dtmc_result)  # type: ignore
    return dtmc_result  # type: ignore



import stormvogel.model
from PIL import Image
from PIL import ImageDraw
from PIL import ImageFont


def create_die_dtmc():
    # Create a new model
    dtmc = stormvogel.model.new_dtmc()

    init = dtmc.get_initial_state()
    init.add_valuation("rolled", 0)

    # From the initial state, add the transition to 6 new states with probability 1/6th.
    init.set_choice(
        [
            (
                1 / 6,
                dtmc.new_state(labels=f"rolled{i + 1}", valuations={"rolled": i + 1}),
            )
            for i in range(6)
        ]
    )

    # we add self loops to all states with no outgoing choices
    dtmc.add_self_loops()

    # test if state removal works
    # dtmc.remove_state(dtmc.get_state_by_id(1), True, True)

    return dtmc


def generate_dice_image(number: int) -> Image.Image:
    """
    Generates a 128x128 PIL image with a question mark for 0
    or a dice face (1-6) with the corresponding number of dots.
    """
    img = Image.new("RGB", (128, 128), "white")
    draw = ImageDraw.Draw(img)

    # Draw border
    draw.rectangle([(5, 5), (123, 123)], outline="black", width=5)  # type: ignore

    if number == 0:
        draw.text((40, 24), "?", font=ImageFont.load_default(80), fill="black")
        return img

    # Dice dot positions relative to a 128x128 grid
    positions = {
        1: [(64, 64)],
        2: [(40, 40), (88, 88)],
        3: [(40, 40), (64, 64), (88, 88)],
        4: [(40, 40), (40, 88), (88, 40), (88, 88)],
        5: [(40, 40), (40, 88), (88, 40), (88, 88), (64, 64)],
        6: [(32, 32), (32, 64), (32, 96), (96, 32), (96, 64), (96, 96)],
    }

    for pos in positions.get(number, []):
        draw.ellipse(
            [(pos[0] - 12, pos[1] - 12), (pos[0] + 12, pos[1] + 12)], fill="black"
        )

    return img


if __name__ == "__main__":
    # Print the resulting model in dot format.

    print(create_die_dtmc().to_dot())



from typing import Tuple
from bidict import bidict

try:
    import stormpy
except ImportError:
    stormpy = None

import stormvogel.model
import stormvogel.stormpy_utils.mapping as mapping


def map_state_labels(m, res):
    """Based on the result of EC elimination, create a new state labeling that can be used for a new model that captures the result.
    Args:
        m: stormpy model
        res (EndComponentEliminatorReturnTypeDouble): EC result
    """
    assert stormpy is not None
    old_nr_columns = m.transition_matrix.nr_columns
    new_nr_columns = res.matrix.nr_columns
    sl = stormpy.StateLabeling(new_nr_columns)

    for s_old in range(old_nr_columns):
        s_new = res.old_to_new_state_mapping[s_old]
        for label in m.labeling.get_labels_of_state(s_old):
            sl.add_label(label)
            sl.add_label_to_state(label, s_new)
    return sl


def map_choice_labels(m_old, m_new, res):
    """Based on the result of EC elimination, create a new choice labeling that can be used for a new model that captures the result.
    Args:
        m_old: old stormpy model
        m_new: new strompy model that is based on res
        res (EndComponentEliminatorReturnTypeDouble): EC result
    """
    assert stormpy is not None
    new_nr_rows = m_new.transition_matrix.nr_rows
    cl = stormpy.storage.ChoiceLabeling(new_nr_rows)
    old_nr_columns = m_old.transition_matrix.nr_columns

    for s_old in range(old_nr_columns):
        s_new = res.old_to_new_state_mapping[s_old]
        old_no_choices = m_old.get_nr_available_actions(s_old)
        new_no_choices = m_new.get_nr_available_actions(s_new)
        if old_no_choices == new_no_choices:
            for action_no in range(old_no_choices):
                old_index = m_old.get_choice_index(s_old, action_no)
                labels = m_old.choice_labeling.get_labels_of_choice(old_index)
                new_index = m_new.get_choice_index(s_new, action_no)
                for label in labels:
                    cl.add_label(label)
                    cl.add_label_to_choice(label, new_index)
    return cl


def simple_ec_elimination(m):
    """Perform End Component elimination on a stormpy model while preserving labels.
    Label sets of merged states are unified.
    Action labels are preserved when possible.
    Args:
        m: stormpy model
    """
    assert stormpy is not None
    # Keep all states, and consider ecs to be possible anywhere in the model
    subsystem = stormpy.BitVector(m.nr_states, True)
    possible_ec_rows = stormpy.BitVector(m.nr_choices, True)
    res = stormpy.eliminate_ECs(
        matrix=m.transition_matrix,
        subsystem=subsystem,
        possible_ecs=possible_ec_rows,
        add_sink_row_states=subsystem,
        add_self_loop_at_sink_states=True,
    )
    new_labels = map_state_labels(m, res)
    components = stormpy.SparseModelComponents(
        transition_matrix=res.matrix, state_labeling=new_labels
    )
    m_new = stormpy.storage.SparseMdp(components)
    components.choice_labeling = map_choice_labels(m, m_new, res)
    m_updated = stormpy.storage.SparseMdp(components)
    return m_updated


def stormvogel_get_maximal_end_components(
    sv_model: stormvogel.model.Model,
) -> list[Tuple[set[int], set[stormvogel.model.Action]]]:
    """Get the maximal end components of this model.
    They are returned as a list of tuples where the first element is a set of state ids, and the second a set of actions.
    """
    assert stormpy is not None
    sp_model = mapping.stormvogel_to_stormpy(sv_model)
    f = choice_mapping(sv_model, sp_model)
    decomposition = stormpy.get_maximal_end_components(sp_model)
    res = []
    for mec in decomposition:
        states = set()
        actions = set()
        for s_id, choices in mec:
            states.add(s_id)
            actions = actions | set(map(lambda x: f.inverse[x], choices))
        res.append((frozenset(states), frozenset(actions)))
    return res


def choice_mapping(sv_model, sp_model):
    """Return a bijective mapping between the stormvogel state-action pairs and the stormvogel model.
    WARNING: This function will be depricated later. It might also be faulty, I don't know :))
    """
    res = bidict({})
    choice_id = 0
    for _, s in sv_model:
        for a in s.available_actions():
            res[s.id, a] = choice_id
            choice_id += 1
    return res



def example_parametric_models_01():
    import stormpy.examples
    import stormpy.examples.files

    path = stormpy.examples.files.prism_pdtmc_die
    prism_program = stormpy.parse_prism_program(path)

    formula_str = "P=? [F s=7 & d=2]"
    properties = stormpy.parse_properties_for_prism_program(formula_str, prism_program)
    model = stormpy.build_parametric_model(prism_program, properties)

    return model



"""Contains the code responsible for model visualization."""

from __future__ import annotations

from collections.abc import Callable
from typing import Any, TYPE_CHECKING

import stormvogel.model
import stormvogel.layout
import stormvogel.result
import stormvogel.html_generation

if TYPE_CHECKING:
    # Type-only imports
    from matplotlib.backend_bases import MouseEvent
    from matplotlib.collections import PathCollection
    from matplotlib.axes import Axes
    import ipywidgets as widgets
    from . import simulator


class VisualizationBase:
    """Base class for visualizing a Stormvogel MDP model.

    This class sets up a visual representation of a Stormvogel model, optionally
    incorporating the result of a model checking operation and a scheduler
    (i.e., a strategy for selecting actions). It constructs an internal graph
    of the model and manages visual layout settings, such as active and
    possible display groups.

    If a scheduler is not explicitly provided but is available in the model
    checking result, it will be used automatically. When a scheduler is set,
    the "scheduled_actions" group is activated in the layout; otherwise, it
    is deactivated.

    Attributes:
        model (stormvogel.model.Model): The MDP model being visualized.
        layout (stormvogel.layout.Layout): The layout configuration for the visualization.
        result (stormvogel.result.Result | None): The result of a model checking operation.
        scheduler (stormvogel.result.Scheduler | None): A scheduler representing a path
            through the model.
        G (ModelGraph): The internal graph structure representing the model.

    Args:
        model (stormvogel.model.Model): The MDP model to visualize.
        layout (stormvogel.layout.Layout, optional): Layout settings for the visualization.
            Defaults to `stormvogel.layout.DEFAULT()`.
        result (stormvogel.result.Result, optional): The result of a model checking
            operation, which may contain a scheduler.
        scheduler (stormvogel.result.Scheduler, optional): An explicit scheduler
            defining actions to take in each state.
    """

    def __init__(
        self,
        model: stormvogel.model.Model,
        layout: stormvogel.layout.Layout = stormvogel.layout.DEFAULT(),
        result: stormvogel.result.Result | None = None,
        scheduler: stormvogel.result.Scheduler | None = None,
    ) -> None:
        self.model = model
        self.layout = layout
        self.result = result
        self.scheduler = scheduler
        # If a scheduler was not set explicitly, but a result was set, then take the scheduler from the results.
        if self.scheduler is None:
            if self.result is not None:
                self.scheduler = self.result.scheduler

        # Set "scheduler" as an active group iff it is present.
        if self.scheduler is not None:
            self.layout.add_active_group("scheduled_actions")
        else:  # Otherwise, disable it
            self.layout.remove_active_group("scheduled_actions")
        self.recreate()

    @staticmethod
    def _und(x: str) -> str:
        """Replace spaces by underscores."""
        return x.replace(" ", "_")

    @staticmethod
    def _random_word(k: int) -> str:
        """Random word of length k"""
        import random
        import string

        return "".join(random.choices(string.ascii_letters, k=k))

    @staticmethod
    def _random_color() -> str:
        """Return a random HEX color."""
        import random

        return "#" + "".join([random.choice("0123456789ABCDEF") for _ in range(6)])

    @staticmethod
    def _blend_colors(c1: str, c2: str, factor: float) -> str:
        """Blend two colors in HEX format. #RRGGBB.
        Args:
            color1 (str): Color 1 in HEX format #RRGGBB
            color2 (str): Color 2 in HEX format #RRGGBB
            factor (float): The fraction of the resulting color that should come from color1.
        """
        r1 = int("0x" + c1[1:3], 0)
        g1 = int("0x" + c1[3:5], 0)
        b1 = int("0x" + c1[5:7], 0)
        r2 = int("0x" + c2[1:3], 0)
        g2 = int("0x" + c2[3:5], 0)
        b2 = int("0x" + c2[5:7], 0)
        r_res = int(factor * r1 + (1 - factor) * r2)
        g_res = int(factor * g1 + (1 - factor) * g2)
        b_res = int(factor * b1 + (1 - factor) * b2)
        return "#" + "".join("%02x" % i for i in [r_res, g_res, b_res])

    def recreate(self):
        """Recreate the ModelGraph and set the edit groups."""
        from .graph import ModelGraph

        self.G = ModelGraph.from_model(
            self.model,
            state_properties=self._create_state_properties,
            action_properties=self._create_action_properties,
            transition_properties=self._create_transition_properties,
        )
        underscored_labels = set(map(self._und, self.model.get_labels()))
        possible_groups = underscored_labels.union(
            {"states", "actions", "scheduled_actions"}
        )
        self.layout.set_possible_groups(possible_groups)

    def _format_number(self, n: stormvogel.model.Value) -> str:
        """Call number_to_string in model.py while accounting for the settings specified in the layout object."""
        if self.layout.layout["numbers"]["visible"] is False:
            return ""
        return stormvogel.model.value_to_string(
            n,
            self.layout.layout["numbers"]["fractions"],
            self.layout.layout["numbers"]["digits"],
            self.layout.layout["numbers"]["denominator_limit"],
        )

    def _format_result(self, s: stormvogel.model.State) -> str:
        """Create a string that shows the result for this state. Starts with newline.
        If results are not enabled, then it returns the empty string."""
        if self.result is None or not self.layout.layout["results"]["show_results"]:
            return ""
        result_of_state = self.result.get_result_of_state(s)
        if result_of_state is None:
            return ""
        return (
            "\n"
            + self.layout.layout["results"]["result_symbol"]
            + " "
            + self._format_number(result_of_state)
        )

    def _format_observations(self, s: stormvogel.model.State) -> str:
        """Create a String that shows the observation for this state (FOR POMDPs).
        Starts with newline."""
        if (
            s.observation is None
            or not self.layout.layout["state_properties"]["show_observations"]
        ):
            return ""
        elif isinstance(s.observation, list):
            string = ""
            for prob, obs in s.observation:
                string += (
                    "\n"
                    + self.layout.layout["state_properties"]["observation_symbol"]
                    + " "
                    + str(obs.format())
                    + ": "
                    + self._format_number(prob)
                )
            return string
        else:
            return (
                "\n"
                + self.layout.layout["state_properties"]["observation_symbol"]
                + " "
                + str(s.observation.format())
            )

    def _group_state(self, s: stormvogel.model.State, default: str) -> str:
        """The user can edit a number of subsets of the states individually, we call these groups.
        This function determines the group of this state. That is, the label of s that has the highest priority, as specified by the user under edit_groups.
        """
        und_labels = set(map(lambda x: self._und(x), s.labels))
        res = list(
            filter(
                lambda x: x in und_labels, self.layout.layout["edit_groups"]["groups"]
            )
        )
        return self._und(res[0]) if res != [] else default

    def _group_action(self, s_id: int, a: stormvogel.model.Action, default: str) -> str:
        """Return the group of this action. Only relevant for scheduling"""
        # Put the action in the group scheduled_actions if appropriate.
        if self.scheduler is None:
            return default

        action = self.scheduler.get_action_at_state(self.model.get_state_by_id(s_id))
        return "scheduled_actions" if a == action else default

    def _format_rewards(
        self, s: stormvogel.model.State, a: stormvogel.model.Action
    ) -> str:
        """Create a string that contains either the state exit reward (if actions are not supported)
        or the reward of taking this action from this state. (if actions ARE supported)
        Starts with newline."""
        if not self.layout.layout["state_properties"]["show_rewards"]:
            return ""
        EMPTY_RES = "\n" + self.layout.layout["state_properties"]["reward_symbol"]
        res = EMPTY_RES
        for reward_model in self.model.rewards:
            if self.model.supports_actions():
                if a in s.available_actions():
                    reward = reward_model.get_state_action_reward(s, a)
                else:
                    reward = None
            else:
                reward = reward_model.get_state_reward(s)
            if reward is not None and not (
                not self.layout.layout["state_properties"]["show_zero_rewards"]
                and reward == 0
            ):
                res += f"\t{reward_model.name}: {self._format_number(reward)}"
        if res == EMPTY_RES:
            return ""
        return res

    def _create_state_properties(self, state: stormvogel.model.State) -> dict:
        """Generates visualization properties for a given state in the model.

        That is, the visual representation of a state, including
        its label, group assignment, color (based on model checking results),
        and other textual annotations like rewards and observations. These
        properties are used when constructing the `ModelGraph` for visualization.

        If result coloring is enabled and a model checking result is available,
        the state's color is interpolated between two configured colors based
        on the state's result value relative to the maximum result.

        Args:
            state (stormvogel.model.State): The model state for which to generate properties.

        Returns:
            dict: A dictionary containing the state's visualization properties with keys:
                - `"label"`: A string representing the state label, including ID (if enabled),
                  labels, rewards, result, and observations.
                - `"group"`: A string identifying the layout group this state belongs to.
                - `"color"`: A blended RGB color string or None, based on result values.
        """
        from fractions import Fraction

        res = self._format_result(state)
        observations = self._format_observations(state)
        rewards = self._format_rewards(state, stormvogel.model.EmptyAction)
        group = self._group_state(state, "states")
        id_label_part = (
            f"{state.id}\n"
            if self.layout.layout["state_properties"]["show_ids"]
            else ""
        )

        color = None

        result_colors = self.layout.layout["results"]["result_colors"]
        if result_colors and self.result is not None:
            result = self.result.get_result_of_state(state)
            max_result = self.result.maximum_result()
            if isinstance(result, (int, float, Fraction)) and isinstance(
                max_result, (int, float, Fraction)
            ):
                color1 = self.layout.layout["results"]["max_result_color"]
                color2 = self.layout.layout["results"]["min_result_color"]
                factor = result / max_result if max_result != 0 else 0
                color = self._blend_colors(color1, color2, float(factor))
        properties = {
            "label": id_label_part
            + ",".join(state.labels)
            + rewards
            + res
            + observations,
            "group": group,
            "color": color,
        }
        return properties

    def _create_action_properties(
        self, state: stormvogel.model.State, action: stormvogel.model.Action
    ) -> dict:
        """Generates visualization properties for a given action in the model.

        This method creates a label for the action, optionally including any
        associated reward, and includes the original `Action` object for use
        in the visualization or interaction.

        Args:
            state (stormvogel.model.State): The state from which the action originates.
            action (stormvogel.model.Action): The action being evaluated.

        Returns:
            dict: A dictionary containing the action's visualization properties:
                - `"label"`: A string combining the action's labels and reward.
                - `"model_action"`: The original `Action` object.
        """
        reward = self._format_rewards(self.model.get_state_by_id(state.id), action)

        properties = {"label": (action.label or "") + reward, "model_action": action}
        return properties

    def _create_transition_properties(self, state, action, next_state) -> dict:
        """Generates visualization properties for a transition between states.

        This method finds the transition probability for a specific state-action–
        next-state triplet and formats it as a label. If the transition exists,
        the formatted probability is included; otherwise, an empty dictionary is returned.

        Args:
            state (stormvogel.model.State): The source state of the transition.
            action (stormvogel.model.Action): The action taken from the source state.
            next_state (stormvogel.model.State): The target state of the transition.

        Returns:
            dict: A dictionary containing the transition's visualization properties:
                - `"label"`: The formatted transition probability, if found.
                  Otherwise, the dictionary will be empty.
        """
        properties = dict()
        transitions = state.get_outgoing_transitions(action)
        if transitions is None:
            return properties
        for prob, target in transitions:
            if next_state.id == target.id:
                properties["label"] = self._format_number(prob)
                return properties
        return properties


class JSVisualization(VisualizationBase):
    """Handles visualization of a Model using a Network from stormvogel.network."""

    EXTRA_PIXELS: int = 20  # To prevent the scroll bar around the Network.

    def __init__(
        self,
        model: stormvogel.model.Model,
        name: str | None = None,
        result: stormvogel.result.Result | None = None,
        scheduler: stormvogel.result.Scheduler | None = None,
        layout: stormvogel.layout.Layout = stormvogel.layout.DEFAULT(),
        output: widgets.Output | None = None,  # type: ignore[name-defined]
        debug_output: widgets.Output | None = None,  # type: ignore[name-defined]
        use_iframe: bool = False,
        do_init_server: bool = True,
        max_states: int = 1000,
        max_physics_states: int = 500,
    ) -> None:
        """Create and show a visualization of a Model using a visjs Network
        Args:
            model (Model): The stormvogel model to be displayed.
            name (str): Used to name the iframe. ONLY SPECIFY IF YOU KNOW WHAT YOU ARE DOING. You should never create two networks with the same name, they might clash.
            result (Result, optional): A result associatied with the model.
                The results are displayed as numbers on a state. Enable the layout editor for options.
                If this result has a scheduler, then the scheduled actions will have a different color etc. based on the layout
            scheduler (Scheduler, optional): The scheduled actions will have a different color etc. based on the layout
                If both result and scheduler are set, then scheduler takes precedence.
            layout (Layout): Layout used for the visualization.
            output (widgets.Output): The output widget in which the network is rendered.
                Whether this widget is also displayed automatically depends on do_display.
            debug_output (widgets.Output): Output widget that can be used to debug interactive features.
            use_iframe (bool): Set to true iff you want to use an iframe. Defaults to False.
            do_init_server (bool): Set to true iff you want to initialize the server. Defaults to True.
            max_states (int): If the model has more states, then the network is not displayed.
            max_physics_states (int): If the model has more states, then physics are disabled.
        """
        import ipywidgets as widgets  # local, heavy
        import os

        super().__init__(model, layout, result, scheduler)

        try:
            self.initial_state_id = model.get_initial_state().id
        except RuntimeError:
            self.initial_state_id = 0

        if output is None:
            self.output = widgets.Output()
        else:
            self.output = output
        if debug_output is None:
            self.debug_output: widgets.Output = widgets.Output()
        else:
            self.debug_output = debug_output

        # vis stuff
        self.name: str = name or self._random_word(10)
        self.use_iframe: bool = use_iframe
        self.max_states: int = max_states
        self.max_physics_states: int = max_physics_states

        self.do_init_server: bool = do_init_server
        self.network_wrapper: str = ""  # Use this for javascript injection.
        if self.use_iframe:
            self.network_wrapper = (
                f"document.getElementById('{self.name}').contentWindow.nw_{self.name}"
            )
        else:
            self.network_wrapper = f"nw_{self.name}"
        self.new_nodes_hidden: bool = False
        # If the user wants to initialize the server, and we are not building the documentation.
        if do_init_server and os.environ.get("DOCUMENTATION", "0").lower() != "1":
            import stormvogel.communication_server  # ensure submodule is loaded

            self.server: stormvogel.communication_server.CommunicationServer = (
                stormvogel.communication_server.initialize_server()
            )
        else:
            self.server = None  # type: ignore[assignment]

    def _generate_node_js(self) -> str:
        """Generate the required js script for node definition"""
        from .graph import NodeType

        node_js = ""
        for node in self.G.nodes():
            node_attr = self.G.nodes[node]
            label = node_attr.get("label", None)
            color = node_attr.get("color", None)
            group = None
            match self.G.nodes[node]["type"]:
                case NodeType.STATE:
                    group = self._group_state(
                        self.model.get_state_by_id(node), "states"
                    )
                case NodeType.ACTION:
                    in_edges = list(self.G.in_edges(node))
                    assert (
                        len(in_edges) == 1
                    ), "An action node should only have a single incoming edge"
                    state, _ = in_edges[0]
                    group = self._group_action(
                        state, self.G.nodes[node]["model_action"], "actions"
                    )
                    # layout_group_color = self.layout.layout["groups"].get(group)
            # if layout_group_color is not None:
            #     color = layout_group_color.get("color", {"background": color}).get(
            #         "background"
            #     )
            #     # HACK: This is necessary for the selection highlighting to work
            #     # and should not be here
            #     color = None
            # TODO ask Ossmoss what this code was supposed to do exactly. It only seems to waste time?
            current = "{ id: " + str(node)
            if label is not None:
                current += f", label: `{label}`"
            if group is not None:
                current += f', group: "{group}"'
            if node in self.layout.layout["positions"]:
                current += (
                    f", x: {self.layout.layout['positions'][node]['x']}, "
                    f"y: {self.layout.layout['positions'][node]['y']}"
                )
            if self.layout.layout["misc"]["explore"] and node != self.initial_state_id:
                current += ", hidden: true"
                current += ", physics: false"
            if color is not None:
                current += f', color: "{color}"'
            current += " },\n"
            node_js += current
        return node_js

    def _generate_edge_js(self) -> str:
        """Generate the required js script for edge definition"""
        from .graph import NodeType

        edge_js = ""
        # preprocess scheduled actions
        scheduled_action_nodes = []
        for node in self.G.nodes:
            if self.G.nodes[node]["type"] == NodeType.ACTION:
                continue
            for _, target in self.G.out_edges(node):
                if self.G.nodes[target]["type"] == NodeType.STATE:
                    continue
                action = self.G.nodes[target]["model_action"]
                group = self._group_action(node, action, "actions")
                if group == "scheduled_actions":
                    scheduled_action_nodes.append(target)

        for from_, to in self.G.edges():
            edge_attr = self.G.edges[(from_, to)]
            # TODO: in order for the layout to have an effect color should be
            # self.layout.layout["edges"]["color"]["color"]
            # however this breaks the highlighing on selection
            color = None
            scheduled_color = self.layout.layout["groups"].get(
                "scheduled_actions", {"color": {"border": color}}
            )["color"]["border"]
            match [self.G.nodes[from_]["type"], self.G.nodes[to]["type"]]:
                case [NodeType.STATE, NodeType.ACTION]:
                    if to in scheduled_action_nodes:
                        color = scheduled_color
                case [NodeType.ACTION, NodeType.STATE]:
                    if from_ in scheduled_action_nodes:
                        color = scheduled_color
            label = edge_attr.get("label", None)
            current = "{ from: " + str(from_) + ", to: " + str(to)
            if label is not None:
                current += f', label: "{label}"'
            if color is not None:
                current += f', color: "{color}"'
            if self.layout.layout["misc"]["explore"]:
                current += ", hidden: true"
                current += ", physics: false"
            current += " },\n"
            edge_js += current
        return edge_js

    def _get_options(self) -> str:
        """Returns the current layout configuration as a JSON-formatted string.

        This method serializes the layout dictionary used for visualization
        into a readable JSON format with indentation for clarity.

        Returns:
            str: A pretty-printed JSON string representing the current layout configuration.
        """
        import json

        return json.dumps(self.layout.layout, indent=2)

    def set_options(self, options: str) -> None:
        """Sets the layout configuration from a JSON-formatted string.

        This method replaces the current layout with a new one defined by the
        given JSON string. It should be called only before visualization is rendered
        (i.e., before calling `show()`), as it reinitializes the layout.

        Args:
            options (str): A JSON-formatted string representing the layout configuration.
        """
        import json

        options_dict = json.loads(options)
        self.layout = stormvogel.layout.Layout(layout_dict=options_dict)

    def generate_html(self) -> str:
        """Generate an html page representing the current state of the `ModelGraph`"""
        return stormvogel.html_generation.generate_html(
            self._generate_node_js(),
            self._generate_edge_js(),
            self._get_options(),
            self.name,
            self.layout.layout["misc"].get("width", 800),
            self.layout.layout["misc"].get("height", 600),
        )

    def generate_iframe(self) -> str:
        """Generate an iframe for the network, using the html."""
        import html as _html

        return f"""
          <iframe
                id="{self.name}"
                width="{self.layout.layout["misc"].get("width", 800) + self.EXTRA_PIXELS}"
                height="{self.layout.layout["misc"].get("height", 600) + self.EXTRA_PIXELS}"
                sandbox="allow-scripts allow-same-origin"
                frameborder="0"
                srcdoc="{_html.escape(self.generate_html())}"
                border:none !important;
                allowfullscreen webkitallowfullscreen mozallowfullscreen
          ></iframe>"""

    def generate_svg(self, width: int = 800) -> str:
        """Generate an svg rendering for the network."""
        from stormvogel.autoscale_svg import autoscale_svg

        js = f"RETURN({self.network_wrapper}.getSvg());"
        res = self.server.result(js)[1:-1]
        unescaped = res.encode("utf-8").decode("unicode_escape")
        scaled = autoscale_svg(unescaped, width)
        return scaled

    def enable_exploration_mode(self, initial_node_id: int):
        """Enables exploration mode starting from a specified initial state.

        This method activates interactive exploration mode in the visualization
        and sets the starting point for exploration to the given state ID.
        `show()` needs to be called after this method is executed to have an effect.

        Args:
            initial_node_id (int): The ID of the state from which exploration should begin.
        """
        self.initial_state_id = initial_node_id
        self.layout.set_value(["misc", "explore"], True)

    def get_positions(self) -> dict[int, dict[str, int]]:
        """Get the current positions of the nodes on the canvas. Returns empty dict if unsuccessful.
        Example result: {0: {"x": 5, "y": 10}}"""
        import json
        import logging

        if self.server is None:
            with self.debug_output:
                logging.warning(
                    "Server not initialized. Could not retrieve position data."
                )
            raise TimeoutError("Server not initialized.")
        try:
            positions: dict = json.loads(
                self.server.result(
                    f"""RETURN({self.network_wrapper}.network.getPositions())"""
                )
            )
            return {int(k): v for k, v in positions.items()}
        except TimeoutError:
            with self.debug_output:
                logging.warning("Timed out. Could not retrieve position data.")
            raise TimeoutError("Timed out. Could not retrieve position data.")

    def show(self, hidden: bool = False) -> None:
        import logging
        import IPython.display as ipd

        with self.output:  # If there was already a rendered network, clear it.
            ipd.clear_output()
        if len(self.model.get_states()) > self.max_states:
            with self.output:
                print(
                    f"This model has more than {self.max_states} states. If you want to proceed, set max_states to a higher value."
                    f"This is to prevent the browser from crashing, be careful."
                )
            return
        if len(self.model.get_states()) > self.max_physics_states:
            with self.output:
                print(
                    f"This model has more than {self.max_physics_states} states. If you want physics, set max_physics_states to a higher value."
                    f"Physics are disabled to prevent the browser from crashing, be careful."
                )
            self.layout.layout["physics"] = False
            self.layout.copy_settings()
        if self.use_iframe:
            iframe = self.generate_iframe()
        else:
            iframe = self.generate_html()
        with self.output:  # Display the iframe within the Output.
            ipd.clear_output()
            ipd.display(ipd.HTML(iframe))
        if not hidden:
            ipd.display(self.output)
        with self.debug_output:
            logging.info("Called show")

    def update(self) -> None:
        """Updates the visualization with the current layout options.

        This method sends updated layout configuration to the frontend visualization
        by injecting JavaScript code. It is typically used to reflect changes made
        to layout settings after the initial rendering.

        Note:
            This should be called after modifying layout properties if the visualization
            has already been shown, to apply those changes interactively.
        """
        import IPython.display as ipd

        js = f"""{self.network_wrapper}.network.setOptions({self._get_options()});"""
        ipd.display(ipd.Javascript(js))

    def set_node_color(self, node_id: int, color: str | None) -> None:
        """Sets the color of a specific node in the visualization.

        This method updates the visual appearance of a node by changing its color
        via JavaScript. It only takes effect once the network has been fully loaded
        in the frontend.

        Args:
            node_id (int): The ID of the node whose color should be changed.
            color (str | None): The color to apply (e.g., "#ff0000" for red).
                If None, the node color is reset or cleared.

        Note:
            This function requires that the visualization is already rendered
            (i.e., `show()` has been called and completed asynchronously).
        """
        import IPython.display as ipd

        if color is None:
            color = "null"
        else:
            color = f'"{color}"'

        js = f"""{self.network_wrapper}.setNodeColor({node_id}, {color});"""
        ipd.display(ipd.Javascript(js))
        ipd.clear_output()

    def highlight_state(self, state_id: int, color: str | None = "red"):
        """Highlights a single state in the model by changing its color.

        This method changes the color of the specified state node in the visualization.
        Pass `None` to reset or clear the highlight.

        Args:
            state_id (int): The ID of the state to highlight.
            color (str | None, optional): The color to use for highlighting (e.g., "red", "#00ff00").
                Defaults to "red".

        Raises:
            AssertionError: If the state ID does not exist in the model graph.
        """
        assert self.G.nodes.get(state_id) is not None, "State id not in ModelGraph"
        self.set_node_color(state_id, color)

    def highlight_action(
        self, state_id: int, action: stormvogel.model.Action, color: str | None = "red"
    ):
        """Highlights a single action in the model by changing its color.

        This method changes the color of the node representing a specific action
        taken from a given state. Pass `None` to remove the highlight.

        Args:
            state_id (int): The ID of the state from which the action originates.
            action (stormvogel.model.Action): The action to highlight.
            color (str | None, optional): The color to use for highlighting. Defaults to "red".

        Warns:
            UserWarning: If the specified (state, action) pair is not found in the model graph.
        """
        import warnings

        try:
            nt_id = self.G.state_action_id_map[(state_id, action)]
            self.set_node_color(nt_id, color)
        except KeyError:
            warnings.warn(
                "Tried to highlight an action that is not present in this model."
            )

    def highlight_state_set(self, state_ids: set[int], color: str | None = "blue"):
        """Highlights a set of states in the model by changing their color.

        Iterates over each state ID in the provided set and applies the given
        color. Pass `None` to clear highlighting for all specified states.

        Args:
            state_ids (set[int]): A set of state IDs to highlight.
            color (str | None, optional): The color to apply. Defaults to "blue".
        """
        for s_id in state_ids:
            self.set_node_color(s_id, color)

    def highlight_action_set(
        self,
        state_action_set: set[tuple[int, stormvogel.model.Action]],
        color: str = "red",
    ):
        """Highlights a set of actions in the model by changing their color.

        Applies the specified color to all (state, action) pairs in the given set.
        Pass `None` as the color to clear the current highlighting.

        Args:
            state_action_set (set[tuple[int, stormvogel.model.Action]]): A set of
                (state ID, action) pairs to highlight.
            color (str, optional): The color to apply. Defaults to "red".
        """
        for s_id, a in state_action_set:
            self.highlight_action(s_id, a, color)

    def highlight_decomposition(
        self,
        decomp: list[tuple[set[int], set[tuple[int, stormvogel.model.Action]]]],
        colors: list[str] | None = None,
    ):
        """Highlight a set of tuples of (states and actions) in the model by changing their color.
        Args:
            decomp: A list of tuples (states, actions)
            colors (optional): A list of colors for the decompositions. Random colors are picked by default.
        """
        for n, v in enumerate(decomp):
            if colors is None:
                color = self._random_color()
            else:
                color = colors[n]
            self.highlight_state_set(v[0], color)
            self.highlight_action_set(v[1], color)

    def clear_highlighting(self):
        """Clear all highlighting that is currently active, returning all states to their original colors."""
        for s_id, _ in self.model:
            self.set_node_color(s_id, None)
        for a_id in self.G.state_action_id_map.values():
            self.set_node_color(a_id, None)

    def highlight_path(
        self,
        path: simulator.Path,  # type: ignore[name-defined]
        color: str,
        delay: float = 1,
        clear: bool = True,
    ) -> None:
        """Highlight the path that is provided as an argument in the model.
        Args:
            path (simulator.Path): The path to highlight.
            color (str | None): The color that the highlighted states should get (in HTML color standard).
                Set to None, in order to clear existing highlights on this path.
            delay (float): If not None, there will be a pause of a specified time before highlighting the next state in the path.
            clear (bool): Clear the highlighting of a state after it was highlighted. Only works if delay is not None.
                This is particularly useful for highlighting paths with loops."""
        from time import sleep

        seq = path.to_state_action_sequence()
        for i, v in enumerate(seq):
            if isinstance(v, stormvogel.model.State):
                self.set_node_color(v.id, color)
                sleep(delay)
                if clear:
                    self.set_node_color(v.id, None)
            elif (
                isinstance(v, stormvogel.model.Action)
                and (seq[i - 1].id, v) in self.G.state_action_id_map
            ):
                node_id = self.G.state_action_id_map[seq[i - 1].id, v]
                self.set_node_color(node_id, color)
                sleep(delay)
                if clear:
                    self.set_node_color(node_id, None)

    def export(self, output_format: str, filename: str = "export") -> None:
        """
        Export the visualization to your preferred output format.
        The appropriate file extension will be added automatically.

        Parameters:
            output_format (str): Desired export format.
            filename (str): Base name for the exported file.

        Supported output formats (not case-sensitive):

            "HTML"    → An interactive .html file (e.g., draggable nodes)
            "IFrame"  → Exports as an <iframe> wrapped HTML in a .html file
            "PDF"     → Exports to .pdf (via conversion from SVG)
            "SVG"     → Exports to .svg vector image
        """
        import pathlib

        output_format = output_format.lower()
        filename_base = pathlib.Path(filename).with_suffix(
            ""
        )  # remove extension if present

        if output_format == "html":
            html_txt = self.generate_html()
            (filename_base.with_suffix(".html")).write_text(html_txt, encoding="utf-8")

        elif output_format == "iframe":
            iframe = self.generate_iframe()
            (filename_base.with_suffix(".html")).write_text(iframe, encoding="utf-8")

        elif output_format == "svg":
            svg = self.generate_svg()
            (filename_base.with_suffix(".svg")).write_text(svg, encoding="utf-8")

        elif output_format == "pdf":
            svg = self.generate_svg()
            import cairosvg  # local, heavy

            cairosvg.svg2pdf(
                bytestring=svg.encode("utf-8"), write_to=filename_base.name + ".pdf"
            )

        elif output_format == "latex":
            svg = self.generate_svg()
            # Create the 'export' folder if it doesn't exist
            export_folder = pathlib.Path(filename_base)
            export_folder.mkdir(parents=True, exist_ok=True)
            pdf_filename = filename_base.with_suffix(".pdf")
            # Convert SVG to PDF
            import cairosvg  # local, heavy

            cairosvg.svg2pdf(
                bytestring=svg.encode("utf-8"),
                write_to=str(export_folder / pdf_filename),
            )

            # Create the LaTeX file
            latex_content = f"""\\documentclass{{article}}
\\usepackage{{graphicx}}
\\begin{{document}}
\\begin{{figure}}[h!]
\\centering
\\includegraphics[width=\\textwidth]{{{pdf_filename.name}}}
\\caption{{Generated using Stormvogel. TODO insert citing instructions}}
\\end{{figure}}
\\end{{document}}
"""
            # Write the LaTeX code to a .tex file
            (export_folder / filename_base.with_suffix(".tex")).write_text(
                latex_content, encoding="utf-8"
            )

        else:
            raise RuntimeError(f"Export format not supported: {output_format}")


class MplVisualization(VisualizationBase):
    """Matplotlib-based visualization for Stormvogel models.

    Extends the base visualization class to render the model, results, and
    scheduler using Matplotlib. Supports interactive features like node
    highlighting and custom hover behavior.

    This class manages figure creation, state and edge highlighting, and
    optionally allows interaction callbacks when hovering over nodes.

    Args:
        model (stormvogel.model.Model): The model to visualize.
        layout (stormvogel.layout.Layout, optional): Layout configuration for
            the visualization. Defaults to `stormvogel.layout.DEFAULT()`.
        result (stormvogel.result.Result | None, optional): The result of a model
            checking operation, which may contain a scheduler. Defaults to None.
        scheduler (stormvogel.result.Scheduler | None, optional): Explicit scheduler
            defining actions to take in each state. Defaults to None.
        title (str | None, optional): Title of the visualization figure. Defaults to None.
        interactive (bool, optional): Whether to enable interactive features such
            as node hover callbacks. Defaults to False.
        hover_node (Callable | None, optional): Callback function invoked when
            hovering over nodes. Receives parameters
            `(PathCollection, PathCollection, MouseEvent, Axes)`. Defaults to None.
    """

    def __init__(
        self,
        model: stormvogel.model.Model,
        layout: stormvogel.layout.Layout = stormvogel.layout.DEFAULT(),
        result: stormvogel.result.Result | None = None,
        scheduler: stormvogel.result.Scheduler | None = None,
        title: str | None = None,
        interactive: bool = False,
        hover_node: (
            Callable[[PathCollection, PathCollection, MouseEvent, Axes], None] | None
        ) = None,
    ):
        super().__init__(model, layout, result, scheduler)
        self.title = title or ""
        self.interactive = interactive
        self.hover_node = hover_node
        self._highlights: dict[int, str] = dict()
        self._edge_highlights: dict[tuple[int, int], str] = dict()
        self._fig = None
        if self.scheduler is not None:
            self.highlight_scheduler(self.scheduler)

    def highlight_state(
        self, state: stormvogel.model.State | int, color: str = "red"
    ) -> None:
        """Highlights a state node in the visualization by setting its color.

        Args:
            state (stormvogel.model.State | int): The state object or state ID to highlight.
            color (str, optional): The color to apply. Defaults to "red".

        Raises:
            AssertionError: If the state node is not present in the model graph.
        """
        if isinstance(state, stormvogel.model.State):
            state = state.id
        node = state
        assert node in self.G.nodes, f"Node {node} not in graph"
        self._highlights[state] = color

    def highlight_action(
        self,
        state: stormvogel.model.State | int,
        action: stormvogel.model.Action,
        color: str = "red",
    ) -> None:
        """Highlights an action node associated with a state by setting its color.

        Args:
            state (stormvogel.model.State | int): The state object or state ID from which the action originates.
            action (stormvogel.model.Action): The action to highlight.
            color (str, optional): The color to apply. Defaults to "red".

        Raises:
            AssertionError: If the state node is not present in the model graph.
        """
        if isinstance(state, stormvogel.model.State):
            state = state.id
        state_node = state
        assert state_node in self.G.nodes, f"Node {state_node} not in graph"
        action_node = self.G.state_action_id_map[state_node, action]
        self._highlights[action_node] = color

    def highlight_edge(self, from_: int, to_: int, color: str = "red") -> None:
        """Highlights an edge between two nodes by setting its color.

        Args:
            from_ (int): The source node ID of the edge.
            to_ (int): The target node ID of the edge.
            color (str, optional): The color to apply. Defaults to "red".
        """
        self._edge_highlights[from_, to_] = color

    def clear_highlighting(self) -> None:
        """Clear all nodes that are marked for highlighting in the visualization"""
        self._highlights.clear()
        self._edge_highlights.clear()

    def highlight_scheduler(self, scheduler: stormvogel.result.Scheduler) -> None:
        """Highlights states, actions, and edges according to the given scheduler.

        Applies a specific highlight color defined by the layout to all states and
        actions specified by the scheduler’s taken actions, as well as the edges connecting them.
        The color is derived from the layout’s configured group colors for scheduled actions.

        Args:
            scheduler (stormvogel.result.Scheduler): The scheduler containing
                state-action mappings to highlight.
        """
        default_color = self.layout.layout["edges"]["color"]["color"]
        color = self.layout.layout["groups"].get(
            "scheduled_actions", {"color": {"border": default_color}}
        )["color"]["border"]
        for state_id, taken_action in scheduler.taken_actions.items():
            self.highlight_state(state_id, color)
            if taken_action == stormvogel.model.EmptyAction:
                continue
            action_node = self.G.state_action_id_map[(state_id, taken_action)]
            self.highlight_action(state_id, taken_action, color)
            self.highlight_edge(state_id, action_node, color)
            for start, end in self.G.out_edges(action_node):
                self.highlight_edge(start, end, color)

    def add_to_ax(
        self,
        ax,
        node_size: int | dict[int, int] = 300,
        node_kwargs: dict[str, Any] | None = None,
        edge_kwargs: dict[str, Any] | None = None,
    ) -> tuple[Any, Any]:
        """Draws the model graph onto a given Matplotlib Axes.

        This method renders nodes and edges of the model graph on the
        provided Matplotlib `ax` object. It uses layout positions, colors from
        the current layout configuration, and any highlights applied to nodes
        or edges. Node sizes can be specified either as a fixed integer or as a
        dictionary mapping node IDs to sizes.

        Args:
            ax (matplotlib.axes.Axes): The Matplotlib axes to draw the graph on.
            node_size (int or dict[int, int], optional): Size(s) of nodes. If an
                int is given, all nodes are drawn with that size. If a dictionary,
                it must provide sizes for all nodes. Defaults to 300.
            node_kwargs (dict[str, Any], optional): Additional keyword arguments
                passed to `nx.draw_networkx_nodes()`. Defaults to None.
            edge_kwargs (dict[str, Any], optional): Additional keyword arguments
                passed to `nx.draw_networkx_edges()`. Defaults to None.

        Returns:
            tuple: A tuple `(nodes, edges)` where `nodes` is the
                `matplotlib.collections.PathCollection` of drawn nodes and `edges`
                is the `matplotlib.collections.LineCollection` of drawn edges.
        """
        import numpy as np
        import networkx as nx
        from .graph import NodeType

        if node_kwargs is None:
            node_kwargs = dict()
        if edge_kwargs is None:
            edge_kwargs = dict()

        if isinstance(node_size, dict):
            assert all(
                [n in node_size for n in self.G.nodes]
            ), "Not all nodes are present in node_size"
        else:
            node_size = {n: node_size for n in self.G.nodes}

        # fetch the colors from the layout
        node_colors = dict()
        for node in self.G.nodes:
            color = "black"
            layout_group_color = None
            match self.G.nodes[node]["type"]:
                case NodeType.STATE:
                    group = self._group_state(
                        self.model.get_state_by_id(node), "states"
                    )
                    layout_group_color = self.layout.layout["groups"].get(group)
                case NodeType.ACTION:
                    in_edges = list(self.G.in_edges(node))
                    assert (
                        len(in_edges) == 1
                    ), "An action node should only have a single incoming edge"
                    state, _ = in_edges[0]
                    group = self._group_action(
                        state, self.G.nodes[node]["model_action"], "actions"
                    )
                    layout_group_color = self.layout.layout["groups"].get(group)
            if layout_group_color is not None:
                color = layout_group_color.get("color", {"background": color}).get(
                    "background"
                )
            node_colors[node] = color

        edge_colors = dict()
        for edge in self.G.edges:
            edge_colors[edge] = self.layout.layout["edges"]["color"]["color"]

        # Now add highlights
        for node, color in self._highlights.items():
            node_colors[node] = color
        for edge, color in self._edge_highlights.items():
            edge_colors[edge] = color

        pos = {
            int(node): np.array((pos["x"], pos["y"]))
            for node, pos in self.layout.layout["positions"].items()
        }
        if len(pos) != len(self.G.nodes):
            pos = nx.random_layout(self.G)
        edges = nx.draw_networkx_edges(
            self.G,
            pos=pos,
            ax=ax,
            edge_color=[edge_colors[e] for e in self.G.edges],  # type: ignore
            **edge_kwargs,
        )
        nodes = nx.draw_networkx_nodes(
            self.G,
            pos=pos,
            ax=ax,
            node_color=[node_colors[n] for n in self.G.nodes],  # type: ignore
            node_size=[node_size[n] for n in self.G.nodes],  # type: ignore
            **node_kwargs,
        )
        return nodes, edges

    def update(
        self,
        node_size: int | dict[int, int] = 300,
        node_kwargs: dict[str, Any] | None = None,
        edge_kwargs: dict[str, Any] | None = None,
    ):
        """Updates or creates the Matplotlib figure displaying the model graph.

        This method sets up the figure size based on layout settings, draws the
        graph nodes and edges using `add_to_ax`, and applies highlights and titles.
        If `interactive` is enabled, it connects a hover event handler to update
        the plot title dynamically when the mouse moves over nodes.

        Args:
            node_size (int or Sequence[int], optional): Size(s) for the nodes. Can be
                a single integer or a sequence of sizes for each node. Defaults to 300.
            node_kwargs (dict[str, Any], optional): Additional keyword arguments
                passed to `nx.draw_networkx_nodes()`. Defaults to None.
            edge_kwargs (dict[str, Any], optional): Additional keyword arguments
                passed to `nx.draw_networkx_edges()`. Defaults to None.

        Returns:
            matplotlib.figure.Figure: The Matplotlib figure object containing the visualization.
        """
        import matplotlib.pyplot as plt

        px = 1 / plt.rcParams["figure.dpi"]
        figsize = (
            self.layout.layout["misc"].get("width", 800) * px,
            self.layout.layout["misc"].get("height", 600) * px,
        )
        if self._fig is None:
            self._fig, ax = plt.subplots(figsize=figsize)
        else:
            w, h = figsize
            self._fig.set_figwidth(w)
            self._fig.set_figheight(h)
            ax = self._fig.gca()
            ax.clear()
        fig = self._fig
        nodes, edges = self.add_to_ax(
            ax,
            node_size=node_size,
            node_kwargs=node_kwargs,
            edge_kwargs=edge_kwargs,
        )
        ax.set_title(self.title)
        node_list = list(self.G.nodes)

        def update_title(ind):
            idx = ind["ind"][0]
            node = node_list[idx]
            node_attr = self.G.nodes[node]
            ax.set_title(f"{node_attr['type'].name}: {node_attr['label']}")

        def hover(event):
            cont, ind = nodes.contains(event)
            if self.hover_node is not None:
                self.hover_node(nodes, edges, event, ax)  # type: ignore
            else:
                if cont:
                    update_title(ind)
                else:
                    ax.set_title(self.title)
            fig.canvas.draw_idle()

        if self.interactive:
            fig.canvas.mpl_connect("motion_notify_event", hover)
        return fig

    def show(self) -> None:
        import matplotlib.pyplot as plt

        self.update()
        plt.show()



"""Interactive layout editor for Stormvogel layouts using ipywidgets.

Implemented using the DictEditor from dict_editor.py, some clarifications:
- Layout stores and manages loading/saving of the layout file, and the schema file.
- DictEditor creates an editor from a schema and a dict to be edited.
- LayoutEditor creates a DictEditor using the schema stored in layouts/schema.json.
    It also processes what should happen when the user edits something in the layout editor or presses the save/load/reload buttons.
"""

import stormvogel.communication_server
import stormvogel.dict_editor
import stormvogel.displayable
import stormvogel.layout
import stormvogel.visualization

import IPython.display as ipd
import ipywidgets as widgets
import logging


class LayoutEditor(stormvogel.displayable.Displayable):
    def __init__(
        self,
        layout: stormvogel.layout.Layout,
        visualization: stormvogel.visualization.JSVisualization | None = None,
        output: widgets.Output | None = None,
        do_display: bool = True,
        debug_output: widgets.Output = widgets.Output(),
    ) -> None:
        """Create an interactive layout editor, according to the schema. Display it using the show() method.

        Args:
            layout (Layout): The layout to be edited.
            visualization (Visualization, optional): A visualization that uses said layout. Defaults to None. Used to update the layout.
            output (widgets.Output, optional): An output widget within which the layout editor should be displayed. Defaults to None.
            do_display (bool, optional): Set to true iff you want the LayoutEditor to display. Defaults to True.
            debug_output (widgets.Output, optional): Debug information is displayed in this output. Leave to default if that doesn't interest you.
        """
        super().__init__(output, do_display, debug_output)
        self.vis: stormvogel.visualization.JSVisualization | None = visualization
        self.layout: stormvogel.layout.Layout = layout
        if self.vis is not None:
            self.vis.show(hidden=True)
        self.loaded: bool = False  # True iff the layout is done loading.
        self.editor = stormvogel.dict_editor.DictEditor(
            schema=self.layout.schema,
            update_dict=self.layout.layout,
            on_update=self.try_update,
            do_display=False,
        )

    def save_node_positions(self):
        """Try to save the positions of the nodes in the graph to the layout.
        The user is informed if this fails."""
        with self.debug_output:
            logging.debug(f"Status of vis {self.vis}")
        if self.vis is not None:
            with self.output:
                if stormvogel.communication_server.server is None:
                    with self.debug_output:
                        logging.info(
                            "Node positions won't be saved because the server is disabled."
                        )
                    with self.output:
                        print(
                            "Node positions won't be saved because the server is disabled."
                        )
                else:
                    try:
                        positions = self.vis.get_positions()
                        with self.debug_output:
                            logging.debug(positions)
                        self.layout.layout["positions"] = positions
                    except TimeoutError:
                        with self.debug_output:
                            logging.warning(
                                "Failed to save node positions in layout file."
                            )
                        with self.output:
                            self.__warn_failed_positions_save()

    def process_save_button(self):
        """Triggered whenever something is changed in the layout editor,
        but only does something if the save button was pressed."""
        self.layout.layout["saving"]["save_button"] = False
        # Also save the node positions.
        self.save_node_positions()
        try:
            self.layout.save(
                self.layout.layout["saving"]["filename"],
                path_relative=self.layout.layout["saving"]["relative_path"],
            )
        except RuntimeError:
            with self.output:
                print("Filename should end in .json")
        except OSError:
            with self.output:
                print(
                    f"Bad or inaccessible path or filename: {self.layout.layout['saving']['filename']}"
                )

    def process_load_button(self):
        """Triggered whenever something is changed in the layout editor,
        but only does something if the load button was pressed."""
        self.layout.layout["saving"]["load_button"] = False
        try:
            self.layout.load(
                self.layout.layout["saving"]["filename"],
                path_relative=self.layout.layout["saving"]["relative_path"],
            )
        except OSError:
            logging.warning("Loaded file does not exist.")
            with self.output:
                print(
                    f"Loaded file does not exist. {self.layout.layout['saving']['filename']}"
                )

    def process_reload_button(self):
        """Triggered whenever something is changed in the layout editor,
        but only does something if the reload button was pressed."""
        self.layout.layout["reload_button"] = False
        with self.debug_output:
            print("Received reload button request.")
            print(self.vis)

    def try_update(self):
        """Process the updates from the layout editor where required."""
        self.layout.copy_settings()
        if not self.loaded:
            return
        save = self.layout.layout["saving"]["save_button"]
        load = self.layout.layout["saving"]["load_button"]
        reload = self.layout.layout["reload_button"]
        if save:
            self.process_save_button()
        if load:
            self.process_load_button()
        if reload:
            self.process_reload_button()
        # The preceeding methods should never call self.vis.show() or self.vis.recreate() since it's already called here.
        if (
            load or reload
        ):  # If we loaded or reloaded, we need to recreate the visualization.
            if self.vis is not None:
                self.vis.recreate()  # This also updates the edit groups as a side effect, so it should be called first.
                self.vis.show()
            self.show()
        elif (
            self.vis is not None
        ):  # Otherwise, just update the layout within the visualization (color changes, for example).
            self.vis.update()

    def show(self) -> None:
        """Display an interactive layout editor, according to the schema."""
        self.loaded = False
        with self.editor.output:  # Clear existing editor.
            ipd.clear_output()
        self.editor = stormvogel.dict_editor.DictEditor(
            schema=self.layout.schema,
            update_dict=self.layout.layout,
            on_update=self.try_update,
            do_display=False,
            output=widgets.Output(),
        )
        self.editor.show()
        box = widgets.VBox(children=[self.editor.output])
        with self.output:
            ipd.clear_output()
            ipd.display(box)
        self.maybe_display_output()
        self.loaded = True

    def __warn_failed_positions_save(self):
        print(
            f"""Could not save the node positions of this graph in {self.layout.layout["saving"]["filename"]}
Sorry for the inconvenience. See 'Implementation details' in docs."""
        )



def example_building_pomdps_01():
    import stormpy
    import stormpy.info

    import stormpy.examples
    import stormpy.examples.files

    import stormpy.pomdp

    # Check support for parameters

    path = stormpy.examples.files.prism_pomdp_maze
    prism_program = stormpy.parse_prism_program(path)
    formula_str = 'P=? [!"bad" U "goal"]'
    properties = stormpy.parse_properties_for_prism_program(formula_str, prism_program)
    # construct the POMDP
    options = stormpy.BuilderOptions([p.raw_formula for p in properties])
    options.set_build_observation_valuations()
    options.set_build_choice_labels()
    pomdp = stormpy.build_sparse_model_with_options(prism_program, options)
    # make its representation canonic.
    pomdp = stormpy.pomdp.make_canonic(pomdp)

    return pomdp


if __name__ == "__main__":
    print(example_building_pomdps_01())



"""Layout stores and manages loading/saving of the layout file, and the schema file.
This schema file may be used by the LayoutEditor to create an editor for this Layout.
The Layout object also provides methods to manipulate the layout and the schema, such as setting node positions from a NetworkX layout.
"""

from typing import Any, Self

import numpy.typing as npt
import stormvogel.rdict

import os
import json
import copy

PACKAGE_ROOT_DIR = os.path.dirname(
    os.path.realpath(__file__)
)  # Get the stormvogel/ directory.


class Layout:
    """Create a Layout object that stores the layout and schema.
    Upon creation, the layout and schema dicts are loaded from "layouts/default.json" and "layouts/schema.json", unless specified otherwise.
    You can load a custom layout file by setting either path or relative_path, or provide a custom layout dict instead.

    Args:
        path (str, optional): Path to a custom layout file.
            Leave to None for a default layout. Defaults to None.
        path_relative (bool, optional): If set to True, then stormvogel will look for a custom layout
            file relative to the current working directory. Defaults to True.
        layout_dict (dict, optional): If set, this dictionary is used as the layout instead of the
            file specified in path. Whenever keys are not present here, their default values from "layouts/default.json" are used.
    """

    def __init__(
        self,
        path: str | None = None,
        path_relative: bool = True,
        layout_dict: dict | None = None,
    ):
        # Open the default layout file.
        with open(os.path.join(PACKAGE_ROOT_DIR, "layouts/default.json")) as f:
            default_str = f.read()
        self.default_dict: dict = json.loads(default_str)

        if layout_dict is None:
            self.load(path, path_relative)
        else:
            self.layout: dict = stormvogel.rdict.merge_dict(
                self.default_dict, layout_dict
            )
            self.load_schema()

    def load_schema(self) -> None:
        """Load in the schema. Used for the layout editor. Stored as self.schema."""
        with open(os.path.join(PACKAGE_ROOT_DIR, "layouts/schema.json")) as f:
            schema_str = f.read()
        self.schema = json.loads(schema_str)

    def load(self, path: str | None = None, path_relative: bool = True) -> None:
        """Load the layout and schema file at the specified path.
        They are stored as self.layout and self.schema respectively."""
        if path is None:
            self.layout: dict = self.default_dict
        else:
            if path_relative:
                complete_path = os.path.join(os.getcwd(), path)
            else:
                complete_path = path
            with open(complete_path) as f:
                parsed_str = f.read()
            parsed_dict = json.loads(parsed_str)
            # Combine the parsed dict with default to fill missing keys as default values.
            self.layout: dict = stormvogel.rdict.merge_dict(
                self.default_dict, parsed_dict
            )
        self.load_schema()

    def add_active_group(self, group: str) -> None:
        """The user can specify which groups of states can be edited seperately. We refer to such groups as active groups.
        Make a group active if it is not already."""
        if group not in self.layout["edit_groups"]["groups"]:
            self.layout["edit_groups"]["groups"].append(group)

    def remove_active_group(self, group: str) -> None:
        """Make a group inactive if it is not already."""
        if group in self.layout["edit_groups"]["groups"]:
            self.layout["edit_groups"]["groups"].remove(group)

    def set_possible_groups(self, groups: set[str]) -> None:
        """Set the groups of states that the user can choose to make active, under edit_groups in the layout editor."""
        self.schema["edit_groups"]["groups"]["__kwargs"]["allowed_tags"] = list(groups)

        # Save changes to the schema. The visualization object will handle putting nodes into the correct groups.
        groups2 = self.layout["edit_groups"]["groups"]
        self.schema[
            "groups"
        ] = {}  # empty the schema groups, to clear existing groups that we may not want
        for g in groups2:
            # For the settings themselves, we need to manually copy everything.
            layout_group_macro = copy.deepcopy(
                self.layout["__fake_macros"]["__group_macro"]
            )
            # Merge the macro with any existing changes.
            existing = self.layout["groups"][g] if g in self.layout["groups"] else {}
            self.layout["groups"][g] = stormvogel.rdict.merge_dict(
                layout_group_macro, existing
            )

            # For the schema, dict_editor already handles macros, so there is no need to do it manually here.
            if g not in self.schema["groups"]:
                self.schema["groups"][g] = {"__use_macro": "__group_macro"}

    def save(self, path: str, path_relative: bool = True) -> None:
        """Save this layout as a json file. Raises runtime error if a filename does not end in json, and OSError if file not found.

        Args:
            path (str): Path to your layout file.
            path_relative (bool, optional): If set to true, then stormvogel will create a custom layout
                file relative to the current working directory. Defaults to True.
        """
        if path[-5:] != ".json":
            raise RuntimeError("File name should end in .json")
        if path_relative:
            complete_path = os.path.join(os.getcwd(), path)
        else:
            complete_path = path
        with open(complete_path, "w") as f:
            json.dump(self.layout, f, indent=2)

    def set_value(self, path: list[str], value: Any) -> None:
        """Set a value in the layout. Also works if a key in the path does not exist."""
        stormvogel.rdict.rset(self.layout, path, value, create_new_keys=True)

    def __str__(self) -> str:
        return json.dumps(self.layout, indent=2)

    def copy_settings(self) -> None:
        """Copy some settings from one place in the layout to another place in the layout.
        They differ because visjs requires for them to be arranged a certain way which is not nice for an editor.
        """
        self.layout["physics"] = self.layout["misc"]["enable_physics"]

    def set_nx_pos(self, pos: dict[int, npt.NDArray], scale: float = 500) -> Self:
        """Apply NetworkX layout positions to this layout and set physics to false.

        Args:
            pos (dict[int, NDArray]): A dictionary of node positions from a NetworkX graph.
            scale (float): Scaling factor for the positions. Defaults to 500.
        """
        self.set_value(["misc", "enable_physics"], False)
        self.set_value(["physics"], False)
        self.set_value(
            ["positions"],
            {
                k: {"x": float(x * scale), "y": float(y * scale)}
                for k, (x, y) in pos.items()
            },
        )
        return self


# Define template layouts.
def DEFAULT():
    return Layout(
        os.path.join(PACKAGE_ROOT_DIR, "layouts/default.json"), path_relative=False
    )


def EXPLORE():
    default = DEFAULT()
    default.layout["misc"]["explore"] = True
    return default


def SV():
    return Layout(
        os.path.join(PACKAGE_ROOT_DIR, "layouts/sv.json"), path_relative=False
    )


def LTS():
    default = DEFAULT()
    default.layout["numbers"]["visible"] = False
    return default



from stormvogel.extensions.ec_elimination import *  # NOQA
from stormvogel.extensions.gifs import *  # NOQA
from stormvogel.extensions.gym_grid import *  # NOQA
from stormvogel.extensions.visual_algos import *  # NOQA
from stormvogel.extensions.gym_sampling import *  # NOQA



"""Create gifs from stormvogel models."""

import stormvogel.simulator as simulator
import stormvogel.model
from typing import Callable
import imageio
from PIL.Image import Image
import os


def render_model_gif(
    model: stormvogel.model.Model,
    state_to_image: Callable[[stormvogel.model.State], Image],
    scheduler: (
        stormvogel.result.Scheduler
        | Callable[[stormvogel.model.State], stormvogel.model.Action]
        | None
    ) = None,
    path: simulator.Path | None = None,
    filename: str = "my_gif",
    max_length: int = 50,
    fps: int = 2,
    loop: int = 0,
) -> str:
    """Render a stormvogel model to a gif, using the gymnasium_scheduler (A map from state numbers to action numbers) to pick an action.
    Leave as None for a random action.

    Args:
        model (Model): stormvogel model.
        state_to_image (Callable[[State], Image]): Function that takes a state and returns an image.
        scheduler (Scheduler | Callable[[State], Action] | None): Scheduler to use for the simulation.
        path (Path | None): Path to use for the simulation. If both scheduler and path are set, path takes presedence.
        filename (str): The gif will be saved as gifs/filename.gif.
        max_length (int): The maximum amount of steps to simulate.
        fps (int): The frames per second of the gif.
        loop (int): The amount of times the gif will loop. 0 means infinite."""
    if path is None:
        if model.supports_actions() and scheduler is not None:
            path = simulator.simulate_path(model, max_length, scheduler)
        else:
            path = simulator.simulate_path(model, max_length)
    frames = [state_to_image(model.get_initial_state())]  # List to store frames

    for i in range(1, min(max_length, len(path))):
        state = path.get_step(i)
        if not isinstance(state, stormvogel.model.State):
            state = state[1]
        frames.append(state_to_image(state))

    os.makedirs("gifs", exist_ok=True)
    # Save frames as a GIF
    imageio.mimsave(
        "gifs/" + filename + ".gif",
        frames,  # type: ignore
        fps=fps,
        loop=loop,
    )  # type: ignore
    return "gifs/" + filename + ".gif"


def embed_gif(filename: str):
    """Hacky way to embed a gif in a jupyter notebook so that it also works with sphinx docs."""
    import IPython.display as ipd

    with open("GIF" + ".html", "w") as f:
        f.write(f'<img src="{filename}">')
    ipd.display(ipd.HTML(filename="GIF" + ".html"))



from stormvogel.examples.car import create_car_mdp  # NOQA
from stormvogel.examples.debugging import create_debugging_mdp  # NOQA
from stormvogel.examples.die import create_die_dtmc, generate_dice_image  # NOQA
from stormvogel.examples.end_components import create_end_components_mdp  # NOQA
from stormvogel.examples.lion import create_lion_mdp  # NOQA
from stormvogel.examples.monty_hall import create_monty_hall_mdp  # NOQA
from stormvogel.examples.monty_hall_pomdp import create_monty_hall_pomdp  # NOQA
from stormvogel.examples.monty_hall_should import create_monty_hall_mdp2  # NOQA
from stormvogel.examples.nuclear_fusion_ctmc import create_nuclear_fusion_ctmc  # NOQA
from stormvogel.examples.stormpy_examples.stormpy_ctmc import (
    example_building_ctmcs_01,  # NOQA
)
from stormvogel.examples.stormpy_examples.stormpy_dtmc import (
    example_building_dtmcs_01,  # NOQA
)
from stormvogel.examples.stormpy_examples.stormpy_ma import (
    example_building_mas_01,  # NOQA
)

from stormvogel.examples.study import create_study_mdp  # NOQA



import stormvogel.model
import re
import stormvogel.parametric as parametric

import json
from typing import Optional, Union, cast

try:
    import stormpy
except ImportError:
    stormpy = None


def value_to_stormpy(
    value: stormvogel.model.Value,
    variables: list["stormpy.pycarl.Variable"],
    model: "stormvogel.model.Model",
) -> "stormpy.pycarl.cln.FactorizedRationalFunction | stormpy.pycarl.Interval | stormvogel.model.Value":
    """converts a stormvogel transition value to a stormpy (pycarl) value"""

    assert stormpy is not None

    def convert_polynomial(
        polynomial: parametric.Polynomial,
    ) -> "stormpy.pycarl.cln.FactorizedPolynomial":
        """helper function for converting polynomials to pycarl polyomials"""
        assert stormpy is not None

        terms = []
        for exponent, coefficient in polynomial.terms.items():
            if coefficient != 0:
                stormpy_term = stormpy.pycarl.cln.Term(
                    stormpy.pycarl.cln.Rational(coefficient)
                )
                assert isinstance(exponent, tuple)
                for index, exp in enumerate(exponent):
                    for i in range(exp):
                        stormpy_term *= variables[
                            [str(var) for var in variables].index(
                                polynomial.variables[index]
                            )
                        ]
                terms.append(stormpy_term)
        polynomial = stormpy.pycarl.cln.Polynomial(terms)
        factorized_polynomial = stormpy.pycarl.cln.FactorizedPolynomial(
            polynomial, stormpy.pycarl.cln.factorization_cache
        )
        return factorized_polynomial

    if model.is_parametric():
        # we have a special case for numbers as they are not just a specific case of a polynomial in stormvogel
        if isinstance(value, stormvogel.model.Number):
            rational = stormpy.pycarl.cln.Rational(float(value))
            polynomial = stormpy.pycarl.cln.Polynomial(rational)
            factorized_polynomial = stormpy.pycarl.cln.FactorizedPolynomial(
                polynomial, stormpy.pycarl.cln.factorization_cache
            )
            factorized_rational = stormpy.pycarl.cln.FactorizedRationalFunction(
                factorized_polynomial
            )
        elif isinstance(value, parametric.RationalFunction):
            factorized_numerator = convert_polynomial(value.numerator)
            factorized_denominator = convert_polynomial(value.denominator)

            # TODO gives segmentation fault
            factorized_rational = stormpy.pycarl.cln.FactorizedRationalFunction(
                factorized_numerator, factorized_denominator
            )
        else:
            assert isinstance(value, parametric.Polynomial)
            factorized_rational = stormpy.pycarl.cln.FactorizedRationalFunction(
                convert_polynomial(value)
            )

        return factorized_rational
    elif model.is_interval_model():
        # in the case of interval models, we convert intervals, and regular values are converted
        # to intervals where the lower and upper value are the same
        if isinstance(value, stormvogel.model.Interval):
            interval = stormpy.pycarl.Interval(value[0], value[1])
        else:
            interval = stormpy.pycarl.Interval(value, value)

        return interval
    else:
        return value


def stormvogel_to_stormpy(
    model: stormvogel.model.Model,
) -> Optional[
    Union[
        "stormpy.storage.SparseDtmc",
        "stormpy.storage.SparseMdp",
        "stormpy.storage.SparseCtmc",
        "stormpy.storage.SparsePomdp",
    ]
]:
    assert stormpy is not None

    def build_matrix(
        model: stormvogel.model.Model,
        choice_labeling: stormpy.storage.ChoiceLabeling | None,
    ) -> stormpy.storage.SparseMatrix:
        """
        Takes a model and creates a stormpy sparsematrix that represents the same choices.
        We also create the choice_labeling by reference simultaneously
        """

        assert stormpy is not None

        # we precompute the following two values
        nondeterministic = model.supports_actions()
        is_parametric = model.is_parametric()
        is_interval = model.is_interval_model()

        # we distinguish between parametric, interval and regular models
        if is_parametric:
            builder = stormpy.ParametricSparseMatrixBuilder(
                rows=0,
                columns=0,
                entries=0,
                force_dimensions=False,
                has_custom_row_grouping=nondeterministic,
                row_groups=0,
            )
        elif is_interval:
            builder = stormpy.IntervalSparseMatrixBuilder(
                rows=0,
                columns=0,
                entries=0,
                force_dimensions=False,
                has_custom_row_grouping=nondeterministic,
                row_groups=0,
            )
        else:
            builder = stormpy.SparseMatrixBuilder(
                rows=0,
                columns=0,
                entries=0,
                force_dimensions=False,
                has_custom_row_grouping=nondeterministic,
                row_groups=0,
            )

        # we build the matrix
        row_index = 0
        for transition in sorted(model.choices.items()):
            if nondeterministic:
                builder.new_row_group(row_index)
            for action in transition[1]:
                action[1].sort_states()
                for tuple in action[1]:
                    val = value_to_stormpy(tuple[0], variables, model)
                    builder.add_next_value(
                        row=row_index,
                        column=model.stormpy_id[tuple[1].id],
                        value=val,
                    )

                # if there is an action then add the label to the choice
                if (
                    not action[0] == stormvogel.model.EmptyAction
                    and choice_labeling is not None
                ):
                    if action[0].label is not None:
                        choice_labeling.add_label_to_choice(action[0].label, row_index)
                row_index += 1

        return builder.build()

    def add_labels(model: stormvogel.model.Model) -> stormpy.storage.StateLabeling:
        """
        Takes a model and creates a state labelling object that determines which states get which labels in the stormpy representation
        """
        assert stormpy is not None

        # we first initialize all labels
        state_labeling = stormpy.storage.StateLabeling(len(model.states.keys()))
        for label in model.get_labels():
            state_labeling.add_label(label)

        # then we assign the labels to the correct states
        for _, state in model:
            for label in state.labels:
                state_labeling.add_label_to_state(label, model.stormpy_id[state.id])

        return state_labeling

    def new_reward_model(
        model: stormvogel.model.Model,
    ) -> dict[str, stormpy.SparseRewardModel]:
        """
        Takes a model and creates a dictionary of all the stormpy representations of reward models
        """
        assert stormpy is not None

        reward_models = {}
        for rewardmodel in model.rewards:
            reward_models[rewardmodel.name] = stormpy.SparseRewardModel(
                optional_state_action_reward_vector=rewardmodel.get_reward_vector()
            )

        return reward_models

    def add_valuations(model: stormvogel.model.Model) -> stormpy.storage.StateValuation:
        """
        Helps to add the valuations to the sparsemodel using a statevaluation object
        """
        assert stormpy is not None

        manager = stormpy.ExpressionManager()
        valuations = stormpy.storage.StateValuationsBuilder()

        # we create all the variable names
        created_vars = set()
        for state in model.states.values():
            for var in sorted(state.valuations.items()):
                name = str(var[0])
                if name not in created_vars:
                    storm_var = manager.create_integer_variable(name)
                    valuations.add_variable(storm_var)
                    created_vars.add(name)

        # we assign the values to the variables in the states
        for state in model.states.values():
            valuations.add_state(
                model.stormpy_id[state.id],
                integer_values=list(state.valuations.values()),
            )

        return valuations.build()

    def map_dtmc(model: stormvogel.model.Model) -> stormpy.storage.SparseDtmc:
        """
        Takes a simple representation of a dtmc as input and outputs a dtmc how it is represented in stormpy
        """
        assert stormpy is not None

        # we first build the SparseMatrix
        matrix = build_matrix(model, None)

        # then we add the state labels
        state_labeling = add_labels(model)

        # then we add the rewards
        reward_models = new_reward_model(model)

        # we add the valuations
        valuations = add_valuations(model)

        # then we build the dtmc
        if model.is_parametric():
            components = stormpy.SparseParametricModelComponents(
                transition_matrix=matrix,
                state_labeling=state_labeling,
                reward_models=reward_models,
            )
            components.state_valuations = valuations
            dtmc = stormpy.storage.SparseParametricDtmc(components)
        elif model.is_interval_model():
            components = stormpy.SparseIntervalModelComponents(
                transition_matrix=matrix,
                state_labeling=state_labeling,
                reward_models=reward_models,
            )
            components.state_valuations = valuations
            dtmc = stormpy.storage.SparseIntervalDtmc(components)
        else:
            components = stormpy.SparseModelComponents(
                transition_matrix=matrix,
                state_labeling=state_labeling,
                reward_models=reward_models,
            )
            components.state_valuations = valuations
            dtmc = stormpy.storage.SparseDtmc(components)

        return dtmc

    def map_mdp(model: stormvogel.model.Model) -> stormpy.storage.SparseMdp:
        """
        Takes a simple representation of an mdp as input and outputs an mdp how it is represented in stormpy
        """
        assert stormpy is not None

        # we determine the number of choices and the labels
        count = 0
        labels = set()
        for _, state in model:
            for action in state.available_actions():
                count += 1
                if action != stormvogel.model.EmptyAction and action.label is not None:
                    labels.add(action.label)

        # we add the labels to the choice labeling object
        choice_labeling = stormpy.storage.ChoiceLabeling(count)
        for label in labels:
            choice_labeling.add_label(str(label))

        # then we create the matrix and simultanuously add the correct labels to the choices
        matrix = build_matrix(model, choice_labeling=choice_labeling)

        # then we add the state labels
        state_labeling = add_labels(model)

        # then we add the rewards
        reward_models = new_reward_model(model)

        # we add the valuations
        valuations = add_valuations(model)

        # then we build the mdp
        if model.is_parametric():
            components = stormpy.SparseParametricModelComponents(
                transition_matrix=matrix,
                state_labeling=state_labeling,
                reward_models=reward_models,
            )
            components.choice_labeling = choice_labeling
            mdp = stormpy.storage.SparseParametricMdp(components)
        elif model.is_interval_model():
            components = stormpy.SparseIntervalModelComponents(
                transition_matrix=matrix,
                state_labeling=state_labeling,
                reward_models=reward_models,
            )
            components.state_valuations = valuations
            components.choice_labeling = choice_labeling
            mdp = stormpy.storage.SparseIntervalMdp(components)
        else:
            components = stormpy.SparseModelComponents(
                transition_matrix=matrix,
                state_labeling=state_labeling,
                reward_models=reward_models,
            )
            components.state_valuations = valuations
            components.choice_labeling = choice_labeling
            mdp = stormpy.storage.SparseMdp(components)

        return mdp

    def map_ctmc(model: stormvogel.model.Model) -> stormpy.storage.SparseCtmc:
        """
        Takes a simple representation of a ctmc as input and outputs a ctmc how it is represented in stormpy
        """
        assert stormpy is not None

        # we first build the SparseMatrix (in stormvogel these are always the rate choices)
        matrix = build_matrix(model, None)

        # then we add the state labels
        state_labeling = add_labels(model)

        # then we add the rewards
        reward_models = new_reward_model(model)

        # we add the valuations
        valuations = add_valuations(model)

        # then we build the ctmc and we add the exit rates if necessary
        if model.is_parametric():
            components = stormpy.SparseParametricModelComponents(
                transition_matrix=matrix,
                state_labeling=state_labeling,
                reward_models=reward_models,
                rate_choices=True,
            )
            components.state_valuations = valuations
            if not model.exit_rates == {} and model.exit_rates is not None:
                components.exit_rates = list(model.exit_rates.values())

            ctmc = stormpy.storage.SparseParametricCtmc(components)
        elif model.is_interval_model():
            components = stormpy.SparseIntervalModelComponents(
                transition_matrix=matrix,
                state_labeling=state_labeling,
                reward_models=reward_models,
                rate_transitions=True,
            )
            components.state_valuations = valuations
            if not model.exit_rates == {} and model.exit_rates is not None:
                components.exit_rates = list(model.exit_rates.values())

            ctmc = stormpy.storage.SparseIntervalCtmc(components)
        else:
            components = stormpy.SparseModelComponents(
                transition_matrix=matrix,
                state_labeling=state_labeling,
                reward_models=reward_models,
                rate_transitions=True,
            )
            components.state_valuations = valuations
            if not model.exit_rates == {} and model.exit_rates is not None:
                components.exit_rates = list(model.exit_rates.values())

            ctmc = stormpy.storage.SparseCtmc(components)

        return ctmc

    def map_pomdp(model: stormvogel.model.Model) -> stormpy.storage.SparsePomdp:
        """
        Takes a simple representation of an pomdp as input and outputs an pomdp how it is represented in stormpy
        """
        assert stormpy is not None

        # Check if the model contains stochastic observation, since we need to do a transformation in that case
        for state in model.states.values():
            if isinstance(state.observation, list):
                raise NotImplementedError(
                    "Stormpy does not support stochastic observations in POMDPs. Please convert the stochastic observations to deterministic ones before converting the model."
                )

        # we determine the number of choices and the labels
        count = 0
        labels = set()
        for state in model.states.values():
            for action in state.available_actions():
                count += 1
                if (
                    not action == stormvogel.model.EmptyAction
                    and action.label is not None
                ):
                    labels.add(action.label)

        # we add the labels to the choice labeling object
        choice_labeling = stormpy.storage.ChoiceLabeling(count)
        for label in labels:
            choice_labeling.add_label(str(label))

        # then we create the matrix and simultanuously add the correct labels to the choices
        matrix = build_matrix(model, choice_labeling=choice_labeling)

        # then we add the state labels
        state_labeling = add_labels(model)

        # then we add the rewards
        reward_models = new_reward_model(model)

        # we add the valuations
        valuations = add_valuations(model)

        # then we build the pomdp
        if model.is_parametric():
            components = stormpy.SparseParametricModelComponents(
                transition_matrix=matrix,
                state_labeling=state_labeling,
                reward_models=reward_models,
            )
            components.state_valuations = valuations
            observations = []
            for state in model.states.values():
                if state.get_observation() is not None:
                    observations.append(state.get_observation().get_observation())
                else:
                    raise RuntimeError(
                        f"State {state.id} does not have an observation. Please assign an observation to each state."
                    )

            components.observability_classes = observations
            components.choice_labeling = choice_labeling
            pomdp = stormpy.storage.SparseParametricPomdp(components)
        elif model.is_interval_model():
            components = stormpy.SparseIntervalModelComponents(
                transition_matrix=matrix,
                state_labeling=state_labeling,
                reward_models=reward_models,
            )
            components.state_valuations = valuations
            observations = []
            for state in model.states.values():
                if state.get_observation() is not None:
                    observations.append(state.get_observation().get_observation())
                else:
                    raise RuntimeError(
                        f"State {state.id} does not have an observation. Please assign an observation to each state."
                    )

            components.observability_classes = observations
            components.choice_labeling = choice_labeling
            pomdp = stormpy.storage.SparseIntervalPomdp(components)
        else:
            components = stormpy.SparseModelComponents(
                transition_matrix=matrix,
                state_labeling=state_labeling,
                reward_models=reward_models,
            )
            components.state_valuations = valuations
            observations = []
            for state in model.states.values():
                if state.get_observation() is None:
                    raise RuntimeError(
                        f"State {state.id} does not have an observation. Please assign an observation to each state."
                    )
                elif isinstance(state.get_observation(), list):
                    raise NotImplementedError(
                        "Stormpy does not support stochastic observations in POMDPs. Please convert the stochastic observations to deterministic ones before converting the model."
                    )
                else:
                    observations.append(state.get_observation().observation)

            components.observability_classes = observations
            components.choice_labeling = choice_labeling
            pomdp = stormpy.storage.SparsePomdp(components)

        return pomdp

    def map_ma(model: stormvogel.model.Model) -> stormpy.storage.SparseMA:
        """
        Takes a simple representation of an ma as input and outputs an ma how it is represented in stormpy
        """
        assert stormpy is not None

        # we determine the number of choices and the labels
        count = 0
        labels = set()
        for state in model.states.values():
            for action in state.available_actions():
                count += 1
                if (
                    not action == stormvogel.model.EmptyAction
                    and action.label is not None
                ):
                    labels.add(action.label)

        # we add the labels to the choice labeling object
        choice_labeling = stormpy.storage.ChoiceLabeling(count)
        for label in labels:
            choice_labeling.add_label(str(label))

        # then we create the matrix and simultanuously add the correct labels to the choices
        matrix = build_matrix(model, choice_labeling=choice_labeling)

        # then we add the state labels
        state_labeling = add_labels(model)

        # then we add the rewards
        reward_models = new_reward_model(model)

        # we add the valuations
        valuations = add_valuations(model)

        # we create the list of markovian state ids
        assert model.markovian_states is not None
        markovian_states_list = [state.id for state in model.markovian_states]
        if isinstance(markovian_states_list, list):
            markovian_states_bitvector = stormpy.storage.BitVector(
                max(markovian_states_list) + 1,
                markovian_states_list,
            )
        else:
            markovian_states_bitvector = stormpy.storage.BitVector(0)

        # then we build the ma
        if model.is_parametric():
            components = stormpy.SparseParametricModelComponents(
                transition_matrix=matrix,
                state_labeling=state_labeling,
                reward_models=reward_models,
                markovian_states=markovian_states_bitvector,
            )
            components.state_valuations = valuations
            if not model.exit_rates == {} and model.exit_rates is not None:
                components.exit_rates = list(model.exit_rates.values())
            else:
                components.exit_rates = []
            components.choice_labeling = choice_labeling
            ma = stormpy.storage.SparseParametricMA(components)
        elif model.is_interval_model():
            components = stormpy.SparseIntervalModelComponents(
                transition_matrix=matrix,
                state_labeling=state_labeling,
                reward_models=reward_models,
                markovian_states=markovian_states_bitvector,
            )
            components.state_valuations = valuations
            if not model.exit_rates == {} and model.exit_rates is not None:
                components.exit_rates = list(model.exit_rates.values())
            else:
                components.exit_rates = []
            components.choice_labeling = choice_labeling
            ma = stormpy.storage.SparseIntervalMA(components)
        else:
            components = stormpy.SparseModelComponents(
                transition_matrix=matrix,
                state_labeling=state_labeling,
                reward_models=reward_models,
                markovian_states=markovian_states_bitvector,
            )
            components.state_valuations = valuations
            if not model.exit_rates == {} and model.exit_rates is not None:
                components.exit_rates = list(model.exit_rates.values())
            else:
                components.exit_rates = []
            components.choice_labeling = choice_labeling
            ma = stormpy.storage.SparseMA(components)

        return ma

    # we throw the neccessary errors first
    if not model.all_states_outgoing_transition():
        raise RuntimeError(
            "This model has states with no outgoing transitions.\nUse the add_self_loops() function to add self loops to all states with no outgoing transition."
        )

    if model.unassigned_variables():
        raise RuntimeError("Each state should have a value for each variable")

    if not model.all_non_init_states_incoming_transition():
        raise RuntimeError(
            "There is more than one state in this model without incoming transitions."
        )

    if model.has_zero_transition() and not model.supports_rates():
        raise RuntimeError(
            "This model has transitions with probability=0. Stormpy assumes that these do not explicitly exist."
        )

    # we make a mapping between stormvogel and stormpy ids in case they are out of order.
    stormpy_id = {}
    for index, stormvogel_id in enumerate(model.states.keys()):
        stormpy_id[stormvogel_id] = index
    model.stormpy_id = stormpy_id

    # we store the pycarl parameters of a model
    stormpy.pycarl.clear_variable_pool()
    variables = []
    for p in model.get_parameters():
        var = stormpy.pycarl.Variable(p)
        variables.append(var)

    # we check the type to handle the model correctly
    if model.type == stormvogel.model.ModelType.DTMC:
        return map_dtmc(model)
    elif model.type == stormvogel.model.ModelType.MDP:
        return map_mdp(model)
    elif model.type == stormvogel.model.ModelType.CTMC:
        return map_ctmc(model)
    elif model.type == stormvogel.model.ModelType.POMDP:
        return map_pomdp(model)
    elif model.type == stormvogel.model.ModelType.MA:
        return map_ma(model)
    else:
        raise NotImplementedError(
            "Converting this type of model to stormpy is not yet supported"
        )


def value_to_stormvogel(value, sparsemodel) -> stormvogel.model.Value:
    """Converts a stormpy transition value to a stormvogel one"""

    assert stormpy is not None

    def is_float(val) -> bool:
        """we check if we can convert a value to a float"""
        try:
            float(val)
            return True
        except (ValueError, TypeError):
            return False

    def convert_polynomial(
        polynomial: stormpy.pycarl.cln.Polynomial,
    ) -> parametric.Polynomial:
        """helper function for converting pycarl polynomials to stormvogel polynomials"""

        # we create the list of variables
        # TODO make this more concise
        var = polynomial.gather_variables()
        var = re.sub(r"{|}", "", str(var))
        var = re.sub(r"<Variable (\w+).*?>", r"\1", var)
        variables_list = [v.strip() for v in var.split(",")]

        # we convert the polynomial to a more suitable list format
        parts = re.split(r"\+(?![^(]*\))", str(polynomial))
        stripped_parts = [part.replace("(", "").replace(")", "") for part in parts]
        term_list = []
        for p in stripped_parts:
            factors = re.split(r"[*^](?![^(]*\))", str(p))
            term_list.append(factors)

        # we initialize the polynomial
        stormvogel_polynomial = parametric.Polynomial(variables_list)

        # and then we add the terms to it
        for term in term_list:
            # we iterate through all terms, variables and exponents
            index_tuple = [0 for i in range(len(variables_list))]
            for i in range(len(term)):
                for j, var in enumerate(variables_list):
                    # we check if there is an exponent or not
                    if term[i] == var:
                        if i < len(term) - 1 and is_float(term[i + 1]):
                            index_tuple[j] = int(term[i + 1])
                        else:
                            index_tuple[j] = 1

            # we check if there is a coefficient at the beginning
            if is_float(term[0]):
                stormvogel_polynomial.add_term(tuple(index_tuple), float(term[0]))
            else:
                stormvogel_polynomial.add_term(tuple(index_tuple), float(1))

        return stormvogel_polynomial

    if sparsemodel.has_parameters:
        # if the model has parameters, all values are rational functions
        regular_form = value.rational_function()
        numerator = regular_form.numerator
        denominator = regular_form.denominator
        converted_numerator = convert_polynomial(numerator)

        # we only return a polynomial if the denominator is 1
        if denominator.is_constant():
            if float(denominator.constant_part()) == 1:
                if numerator.is_constant():
                    return float(numerator.constant_part())
                else:
                    return converted_numerator
        converted_denominator = convert_polynomial(denominator)
        stormvogel_rational_function = parametric.RationalFunction(
            converted_numerator, converted_denominator
        )
        return stormvogel_rational_function
    else:
        # we check if our value is an interval
        if isinstance(value, stormpy.pycarl.Interval):
            # if lower and upper are the same, we return a singular value
            lower = float(value.lower())
            upper = float(value.upper())
            if lower == upper:
                return lower

            return stormvogel.model.Interval(lower, upper)

        # if our function is just a rational number we return a float:
        return float(value)


def stormpy_to_stormvogel(
    sparsemodel: Union[
        "stormpy.storage.SparseDtmc",
        "stormpy.storage.SparseMdp",
        "stormpy.storage.SparseCtmc",
        "stormpy.storage.SparsePomdp",
        "stormpy.storage.SparseMA",
    ],
) -> stormvogel.model.Model | None:
    assert stormpy is not None

    def add_states(
        model: stormvogel.model.Model,
        sparsemodel: (
            stormpy.storage.SparseDtmc
            | stormpy.storage.SparseMdp
            | stormpy.storage.SparseCtmc
            | stormpy.storage.SparsePomdp
            | stormpy.storage.SparseMA
        ),
    ):
        """
        helper function to add the states from the sparsemodel to the model
        """
        model.new_state()
        for state in sparsemodel.states:
            if state.id == 0:
                for label in state.labels:
                    model.get_state_by_id(0).add_label(label)
            if state.id > 0:
                model.new_state(labels=list(state.labels))

    def new_reward_model(
        model: stormvogel.model.Model,
        sparsemodel: (
            stormpy.storage.SparseDtmc
            | stormpy.storage.SparseMdp
            | stormpy.storage.SparseCtmc
            | stormpy.storage.SparsePomdp
            | stormpy.storage.SparseMA
        ),
    ):
        """
        adds the rewards from the sparsemodel to either the states or the state action pairs of the model
        """
        for reward_model_name in sparsemodel.reward_models:
            rewards = sparsemodel.get_reward_model(reward_model_name)
            rewardmodel = model.new_reward_model(reward_model_name)
            get_reward_vector = (
                rewards.state_action_rewards
                if rewards.has_state_action_rewards
                else rewards.state_rewards
            )

            rewardmodel.set_from_rewards_vector(get_reward_vector)

    def add_valuations(
        model: stormvogel.model.Model,
        sparsemodel: (
            stormpy.storage.SparseDtmc
            | stormpy.storage.SparseMdp
            | stormpy.storage.SparseCtmc
            | stormpy.storage.SparsePomdp
            | stormpy.storage.SparseMA
        ),
    ):
        """
        adds the valuations from the sparsemodel to the states of the model
        """
        if sparsemodel.has_state_valuations():
            valuations = sparsemodel.state_valuations

            for state_id, state in model.states.items():
                v = json.loads(str(valuations.get_json(state_id)))
                if v is not None:
                    state.valuations = v

    def map_dtmc(sparsedtmc: stormpy.storage.SparseDtmc) -> stormvogel.model.Model:
        """
        Takes a dtmc stormpy representation as input and outputs a simple stormvogel representation
        """

        # we create the model
        model = stormvogel.model.new_dtmc(create_initial_state=False)

        # we add the states
        add_states(model, sparsedtmc)

        # we add the transitions
        matrix = sparsedtmc.transition_matrix
        for state in sparsedtmc.states:
            row = matrix.get_row(state.id)
            choiceshorthand = [
                (
                    value_to_stormvogel(x.value(), sparsedtmc),
                    model.get_state_by_id(x.column),
                )
                for x in row
            ]
            choices = stormvogel.model.choice_from_shorthand(
                cast(
                    list[tuple[stormvogel.model.Value, stormvogel.model.State]],
                    choiceshorthand,
                ),
                model,
            )
            model.set_choice(model.get_state_by_id(state.id), choices)

        # we add the valuations
        add_valuations(model, sparsedtmc)

        # we add self loops to all states with no outgoing transition
        model.add_self_loops()

        # we add the reward models to the states
        new_reward_model(model, sparsedtmc)

        return model

    def map_mdp(sparsemdp: stormpy.storage.SparseMdp) -> stormvogel.model.Model:
        """
        Takes a mdp stormpy representation as input and outputs a simple stormvogel representation
        """

        # we create the model
        model = stormvogel.model.new_mdp(create_initial_state=False)

        # we add the states
        add_states(model, sparsemdp)

        # we add the transitions
        matrix = sparsemdp.transition_matrix
        for index, state in enumerate(sparsemdp.states):
            row_group_start = matrix.get_row_group_start(index)
            row_group_end = matrix.get_row_group_end(index)

            # within a row group we add for each action the transitions
            choice = dict()
            for i in range(row_group_start, row_group_end):
                row = matrix.get_row(i)

                if sparsemdp.has_choice_labeling():
                    labels = sparsemdp.choice_labeling.get_labels_of_choice(i)
                    actionlabel = ",".join(sorted(labels)) if labels else None
                else:
                    actionlabel = str(i)

                action = model.action(actionlabel)
                branch = [
                    (
                        value_to_stormvogel(x.value(), sparsemdp),
                        model.get_state_by_id(x.column),
                    )
                    for x in row
                ]
                choice[action] = stormvogel.model.Branches(
                    cast(
                        list[tuple[stormvogel.model.Value, stormvogel.model.State]],
                        branch,
                    )
                )
                model.set_choice(model.get_state_by_id(state.id), choice)

        # we add self loops to all states with no outgoing transitions
        model.add_self_loops()

        # we add the reward models to the state action pairs
        new_reward_model(model, sparsemdp)

        # we add the valuations
        add_valuations(model, sparsemdp)

        return model

    def map_ctmc(sparsectmc: stormpy.storage.SparseCtmc) -> stormvogel.model.Model:
        """
        Takes a ctmc stormpy representation as input and outputs a simple stormvogel representation
        """

        # we create the model
        model = stormvogel.model.new_ctmc(create_initial_state=False)

        # we add the states
        add_states(model, sparsectmc)

        # we add the transitions
        matrix = sparsectmc.transition_matrix
        for state in sparsectmc.states:
            row = matrix.get_row(state.id)
            choiceshorthand = [
                (
                    value_to_stormvogel(x.value(), sparsectmc),
                    model.get_state_by_id(x.column),
                )
                for x in row
            ]
            choices = stormvogel.model.choice_from_shorthand(
                cast(
                    list[tuple[stormvogel.model.Value, stormvogel.model.State]],
                    choiceshorthand,
                ),
                model,
            )
            model.set_choice(model.get_state_by_id(state.id), choices)

        # we add self loops to all states with no outgoing transitions
        model.add_self_loops()

        # we add the reward models to the states
        new_reward_model(model, sparsectmc)

        # we add the valuations
        add_valuations(model, sparsectmc)

        # we set the correct exit rates
        for state in model.states.items():
            model.set_rate(state[1], sparsectmc.exit_rates[state[1].id])

        return model

    def map_pomdp(sparsepomdp: stormpy.storage.SparsePomdp) -> stormvogel.model.Model:
        """
        Takes a pomdp stormpy representation as input and outputs a simple stormvogel representation
        """

        # we create the model
        model = stormvogel.model.new_pomdp(create_initial_state=False)

        # we add the states
        add_states(model, sparsepomdp)

        # we add the transitions
        matrix = sparsepomdp.transition_matrix
        for index, state in enumerate(sparsepomdp.states):
            row_group_start = matrix.get_row_group_start(index)
            row_group_end = matrix.get_row_group_end(index)

            # within a row group we add for each action the transitions
            choice = dict()
            for i in range(row_group_start, row_group_end):
                row = matrix.get_row(i)

                if sparsepomdp.has_choice_labeling():
                    labels = sparsepomdp.choice_labeling.get_labels_of_choice(i)
                    actionlabel = ",".join(sorted(labels)) if labels else None
                else:
                    actionlabel = str(i)

                action = model.action(actionlabel)
                branch = [
                    (
                        value_to_stormvogel(x.value(), sparsepomdp),
                        model.get_state_by_id(x.column),
                    )
                    for x in row
                ]
                choice[action] = stormvogel.model.Branches(
                    cast(
                        list[tuple[stormvogel.model.Value, stormvogel.model.State]],
                        branch,
                    )
                )
                model.set_choice(model.get_state_by_id(state.id), choice)

        # we add self loops to all states with no outgoing transitions
        model.add_self_loops()

        # we add the reward models to the state action pairs
        new_reward_model(model, sparsepomdp)

        # we add the valuations
        add_valuations(model, sparsepomdp)

        # we add the observations:
        for state in model.states.values():
            state.set_observation(
                model.observation(sparsepomdp.get_observation(state.id))
            )

        return model

    def map_ma(sparsema: stormpy.storage.SparseMA) -> stormvogel.model.Model:
        """
        Takes a ma stormpy representation as input and outputs a simple stormvogel representation
        """

        # we create the model
        model = stormvogel.model.new_ma(create_initial_state=False)

        # we add the states
        add_states(model, sparsema)

        # we add the transitions
        matrix = sparsema.transition_matrix
        for index, state in enumerate(sparsema.states):
            row_group_start = matrix.get_row_group_start(index)
            row_group_end = matrix.get_row_group_end(index)

            # within a row group we add for each action the transitions
            choice = dict()
            for i in range(row_group_start, row_group_end):
                row = matrix.get_row(i)

                if sparsema.has_choice_labeling():
                    labels = sparsema.choice_labeling.get_labels_of_choice(i)
                    actionlabel = ",".join(sorted(labels)) if labels else None
                else:
                    actionlabel = str(i)

                action = model.action(actionlabel)
                branch = [
                    (
                        value_to_stormvogel(x.value(), sparsema),
                        model.get_state_by_id(x.column),
                    )
                    for x in row
                ]
                choice[action] = stormvogel.model.Branches(
                    cast(
                        list[tuple[stormvogel.model.Value, stormvogel.model.State]],
                        branch,
                    )
                )
                model.set_choice(model.get_state_by_id(state.id), choice)

        # we add self loops to all states with no outgoing transitions
        model.add_self_loops()

        # we add the reward models to the state action pairs
        new_reward_model(model, sparsema)

        # we add the valuations
        add_valuations(model, sparsema)

        # we set the correct exit rates
        for state in model.states.items():
            model.set_rate(state[1], sparsema.exit_rates[state[1].id])

        # we set the markovian states
        for state_id in list(sparsema.markovian_states):
            model.add_markovian_state(model.get_state_by_id(state_id))

        return model

    # we check the type to handle the sparse model correctly
    if sparsemodel.model_type.name == "DTMC":
        return map_dtmc(sparsemodel)
    elif sparsemodel.model_type.name == "MDP":
        return map_mdp(sparsemodel)
    elif sparsemodel.model_type.name == "CTMC":
        return map_ctmc(sparsemodel)
    elif sparsemodel.model_type.name == "POMDP":
        return map_pomdp(sparsemodel)
    elif sparsemodel.model_type.name == "MA":
        return map_ma(sparsemodel)
    else:
        raise NotImplementedError(
            "Converting this type of model to stormvogel is not yet supported"
        )


def from_prism(prism_code="stormpy.storage.storage.PrismProgram"):
    """Create a model from prism."""

    assert stormpy is not None
    return stormpy_to_stormvogel(stormpy.build_model(prism_code))



"""Functionality for handling nested dictionaries easily."""

import copy
from functools import reduce
from typing import Any


def rget(d: dict, path: list) -> Any:
    """Recursively get dict value. Throws KeyError if (nested) key not present.

    Example: rget(d, ['a', 'b', 'c']) is equivalent to d['a']['b']['c']"""
    return reduce(lambda c, k: c.__getitem__(k), path, d)


def rset(d: dict, path: list[str], value: Any, create_new_keys: bool = False) -> dict:
    """Recursively set dict value.

    Args:
        d (dict): dictionary to modify.
        path (list[str]): A list of keys that lead to the value you want to set. The first key is the first key in the path, the second key is the second key in the path, etc.
        value (Any): Target value.
        create_new_keys (bool): If a key that is on the path does not exist yet, create it. Defaults to False.

    Example: rset(d, ['a', 'b', 'c'], 5) is equivalent to d['a']['b']['c'] = 5
    """
    if len(path) == 0:
        return d

    def __rset(d: dict, path: list, value: Any):
        first = path.pop(0)
        if create_new_keys and first not in d:
            d[first] = {}
        if len(path) == 0:
            d[first] = value
        else:
            __rset(d[first], path, value)

    __rset(d, copy.deepcopy(path), value)
    return d


def merge_dict(dict1: dict, dict2: dict) -> dict:
    """Merge two nested dictionaries recursively. Note that dict1 is modified by reference and also returned.

    Args:
        dict1 (dict):
        dict2 (dict):

    In general, dict2 gets priority!
    If dict2 has a value that dict1 does not have, then the value in dict2 is chosen.
    If dict1 and dict2 have the same key, and both are VALUES, then dict2 is chosen.
    If dict1 and dict2 have the same key, and both are DICTIONARIES, then the two dictionaries are merged recursively.
    If dict1 has a DICTIONARY and dict2 has a VALUE with the same key, then dict1 gets priority.

    Taken from StackOverflow user Anatoliy R on July 2 2024.
    https://stackoverflow.com/questions/43797333/how-to-merge-two-nested-dict-in-python
    """
    for key, val in dict1.items():
        if isinstance(val, dict):
            if key in dict2 and type(dict2[key] == dict):
                merge_dict(dict1[key], dict2[key])
        else:
            if key in dict2:
                dict1[key] = dict2[key]

    for key, val in dict2.items():
        if key not in dict1:
            dict1[key] = val

    return dict1



import stormvogel.model


def create_monty_hall_mdp():
    mdp = stormvogel.model.new_mdp()

    init = mdp.get_initial_state()

    # first choose car position
    init.set_choice(
        [(1 / 3, mdp.new_state("carchosen", {"car_pos": i})) for i in range(3)]
    )

    # we choose a door in each case
    for s in mdp.get_states_with_label("carchosen"):
        s.set_choice(
            [
                (
                    mdp.action(f"open{i}"),
                    mdp.new_state("open", s.valuations | {"chosen_pos": i}),
                )
                for i in range(3)
            ]
        )

    # the other goat is revealed
    for s in mdp.get_states_with_label("open"):
        car_pos = s.valuations["car_pos"]
        chosen_pos = s.valuations["chosen_pos"]
        assert isinstance(car_pos, int) and isinstance(chosen_pos, int)
        other_pos = {0, 1, 2} - {car_pos, chosen_pos}
        s.set_choice(
            [
                (
                    1 / len(other_pos),
                    mdp.new_state("goatrevealed", s.valuations | {"reveal_pos": i}),
                )
                for i in other_pos
            ]
        )

    # we must choose whether we want to switch
    for s in mdp.get_states_with_label("goatrevealed"):
        car_pos = s.valuations["car_pos"]
        chosen_pos = s.valuations["chosen_pos"]
        reveal_pos = s.valuations["reveal_pos"]
        assert isinstance(reveal_pos, int) and isinstance(chosen_pos, int)
        other_pos = list({0, 1, 2} - {reveal_pos, chosen_pos})[0]
        s.set_choice(
            [
                (
                    mdp.action("stay"),
                    mdp.new_state(
                        ["done"] + (["target"] if chosen_pos == car_pos else ["lost"]),
                        s.valuations | {"chosen_pos": chosen_pos},
                    ),
                ),
                (
                    mdp.action("switch"),
                    mdp.new_state(
                        ["done"] + (["target"] if other_pos == car_pos else ["lost"]),
                        s.valuations | {"chosen_pos": other_pos},
                    ),
                ),
            ]
        )

    # we add self loops to all states with no outgoing choices
    mdp.add_self_loops()

    # we set the value -1 to all unassigned variables in the states
    mdp.add_valuation_at_remaining_states(value=-1)

    return mdp


if __name__ == "__main__":
    # Print the resulting model in dot format.

    print(create_monty_hall_mdp().to_dot())



"""Shorter api for showing a model."""

from typing import Callable, Any, TYPE_CHECKING
import stormvogel.model
from stormvogel.layout import Layout, DEFAULT, SV
import stormvogel.result
from stormvogel.visualization import JSVisualization, MplVisualization

if TYPE_CHECKING:
    import ipywidgets as widgets
    from stormvogel.graph import ModelGraph


def show(
    model: stormvogel.model.Model,
    result: stormvogel.result.Result | None = None,
    engine: str = "js",
    pos_function: Callable[["ModelGraph"], dict[int, Any]] | None = None,
    pos_function_scaling: int = 750,
    scheduler: stormvogel.result.Scheduler | None = None,
    layout: Layout | None = None,
    show_editor: bool = False,
    debug_output: "widgets.Output | None" = None,
    use_iframe: bool = False,
    do_init_server: bool = True,
    max_states: int = 1000,
    max_physics_states: int = 500,
) -> JSVisualization | MplVisualization | None:
    """Create and show a visualization of a Model using a visjs Network

    Args:
        model (Model): The stormvogel model to be displayed.
        engine (str): The engine that should be used for the visualization.
            Can be either "js" for the interactive html/JavaScript visualization, or "mpl" for matplotlib.
        pos_function (Callable | None): Function that takes a ModelGraph and maps it to a dictionary of node positions.
            It is often useful to import these from networkx, see https://networkx.org/documentation/stable/_modules/networkx/drawing/layout.html for some examples.
            In particular, nx.nx.bfs_layout seems to work great for models with a directed acyclic graph structure.
            Defaults to None.
        pos_function_scaling (int): Scaling factor for the positions when using networkx positions. Defaults to 500.
        result (Result, optional): A result associatied with the model.
            The results are displayed as numbers on a state. Enable the layout editor for options.
            If this result has a scheduler, then the scheduled actions will have a different color etc. based on the layout
        scheduler (Scheduler, optional): The scheduled actions will have a different color etc. based on the layout
            If both result and scheduler are set, then scheduler takes precedence.
        layout (Layout): Layout used for the visualization.
        show_editor (bool): For interactive visualizaiton. Show an interactive layout editor.
        debug_output (widgets.Output): For interactive visualization. Output widget that can be used to debug interactive features.
        use_iframe(bool): For interactive visualziation. Wrap the generated html inside of an IFrame.
            In some environments, the visualization works better with this enabled.
        do_init_server(bool): For interactive visualization. Initialize a local server that is used for communication between Javascript and Python.
            If this is set to False, then exporting network node positions and svg/pdf/latex is impossible.
        max_states (int): If the model has more states, then the network is not displayed.
        max_physics_states (int): If the model has more states, then physics are disabled.

    Returns:
        JSVisualization or MlpVisualzition object.
    """
    import ipywidgets as widgets
    import IPython.display as ipd

    if layout is None:
        layout = DEFAULT()

    # Use networkx positions if the user wants it.
    if pos_function:
        from stormvogel.graph import ModelGraph

        G = ModelGraph.from_model(model)
        try:
            pos = pos_function(G)
            layout = layout.set_nx_pos(pos, scale=pos_function_scaling)
        finally:
            pass

    if engine == "js":
        vis = JSVisualization(
            model=model,
            result=result,
            scheduler=scheduler,
            layout=layout,
            debug_output=debug_output or widgets.Output(),
            do_init_server=do_init_server,
            use_iframe=use_iframe,
            max_states=max_states,
            max_physics_states=max_physics_states,
        )
        if show_editor:
            import stormvogel.layout_editor

            e = stormvogel.layout_editor.LayoutEditor(
                layout,
                vis,
                do_display=False,
                debug_output=debug_output or widgets.Output(),
            )
            e.show()
            box = widgets.HBox(children=[vis.output, e.output])
            ipd.display(box)
        else:  # Unfortunately, the sphinx docs only work if we save the html as a file and embed.
            if use_iframe:
                iframe = vis.generate_iframe()
            else:
                iframe = vis.generate_html()
            with open("model.html", "w") as f:
                f.write(iframe)
            ipd.display(ipd.HTML(filename="model.html"))
        return vis
    elif engine == "mpl":
        vis = MplVisualization(
            model=model, result=result, scheduler=scheduler, layout=layout
        )
        vis.show()
        return vis
    else:
        print(f"Unkown engine: {engine}. Choose 'js' or 'mlp'.")


def show_bird() -> JSVisualization:
    """Show a simple model with a bird state."""
    m = stormvogel.model.new_dtmc(create_initial_state=False)
    m.new_state("🐦")
    m.add_self_loops()
    vis = show(m, show_editor=False, do_init_server=False, layout=SV())
    assert isinstance(vis, JSVisualization)
    return vis



import stormvogel.model


def create_study_mdp():
    mdp = stormvogel.model.new_mdp()

    init = mdp.get_initial_state()
    study = mdp.action("study")
    not_study = mdp.action("don't study")

    pass_test = mdp.new_state("pass test")
    fail_test = mdp.new_state("fail test")
    end = mdp.new_state("end")

    init.set_choice(
        {
            study: [(9 / 10, pass_test), (1 / 10, fail_test)],
            not_study: [(4 / 10, pass_test), (6 / 10, fail_test)],
        }
    )

    pass_test.set_choice([(1, end)])
    fail_test.set_choice([(1, end)])

    mdp.add_self_loops()

    reward_model = mdp.new_reward_model("R")
    reward_model.set_state_action_reward(pass_test, stormvogel.model.EmptyAction, 100)
    reward_model.set_state_action_reward(fail_test, stormvogel.model.EmptyAction, 0)
    reward_model.set_state_action_reward(init, not_study, 15)
    reward_model.set_state_action_reward(init, study, 0)
    reward_model.set_unset_rewards(0)

    return mdp



from stormvogel import parametric
from stormvogel import model, bird


def create_knuth_yao_pmc():
    # we first make polynomials 'x' and '1-x'
    x = parametric.Polynomial(["x"])
    x.add_term((1,), 1)

    invx = parametric.Polynomial(["x"])
    invx.add_term((1,), -1)
    invx.add_term((0,), 1)

    # we build the knuth yao dice using the bird model builder
    initial_state = bird.State(s=0)

    def delta(s: bird.State):
        match s.s:
            case 0:
                return [(x, bird.State(s=1)), (invx, bird.State(s=2))]
            case 1:
                return [(x, bird.State(s=3)), (invx, bird.State(s=4))]
            case 2:
                return [(x, bird.State(s=5)), (invx, bird.State(s=6))]
            case 3:
                return [(x, bird.State(s=1)), (invx, bird.State(s=7, d=1))]
            case 4:
                return [
                    (x, bird.State(s=7, d=2)),
                    (invx, bird.State(s=7, d=3)),
                ]
            case 5:
                return [
                    (x, bird.State(s=7, d=4)),
                    (invx, bird.State(s=7, d=5)),
                ]
            case 6:
                return [(x, bird.State(s=2)), (invx, bird.State(s=7, d=6))]
            case 7:
                return [(1, s)]

    # we add labels to the final states
    def labels(s: bird.State) -> str | None:
        if s.s == 7:
            return f"rolled{str(s.d)}"

    return bird.build_bird(
        delta=delta,
        labels=labels,
        init=initial_state,
        modeltype=model.ModelType.DTMC,
    )


if __name__ == "__main__":
    print(create_knuth_yao_pmc())



from stormvogel import bird
from stormvogel.model import ModelType


def create_end_components_mdp():
    init = bird.State(x=["init"])

    def available_actions(s: bird.State):
        if "init" in s.x:
            return ["one", "two"]
        elif "mec1" in s.x or "mec2" in s.x:
            return ["one", "two"]
        return [""]

    def delta(s: bird.State, a: bird.Action):
        if "init" in s.x and a == "one":
            return [(0.5, bird.State(x=["mec1"])), (0.5, bird.State(x=["mec2"]))]
        elif "mec1" in s.x:
            return [(1, bird.State(x=["mec2"]))]
        elif "mec2" in s.x:
            return [(1, bird.State(x=["mec1"]))]
        elif "init" in s.x and a == "two":
            return [(1, bird.State(x=["mec1"]))]
        return [(1, s)]

    def labels(s):
        return s.x

    mdp = bird.build_bird(
        delta=delta,
        init=init,
        available_actions=available_actions,
        labels=labels,
        modeltype=ModelType.MDP,
    )
    return mdp



import stormvogel.model


def create_car_mdp():
    car = stormvogel.model.new_mdp()
    gs = car.get_initial_state()
    gs.add_label("green light")
    gs.add_label("still")
    rs = car.new_state(["red light", "still"])
    gm = car.new_state(["green light", "moving"])
    rm = car.new_state(["red light", "moving"])
    accident = car.new_state("accident")

    drive = car.new_action("accelerate")
    brake = car.new_action("brake")
    wait = car.new_action("wait")

    gs.set_choice([(drive, gm), (wait, rs)])
    rs.set_choice([(drive, rm), (wait, gs)])
    gm.set_choice([(brake, gs), (wait, rm)])
    rm.set_choice([(brake, rs), (wait, accident)])
    car.add_self_loops()
    return car



import stormvogel.model


def build_property_string(model: stormvogel.model.Model):
    """Lets the user build a property string using a widget."""
    import IPython.display as ipd
    import ipywidgets as widgets
    from stormvogel.dict_editor import DictEditor

    # The first part of the property builder: choose P, R or S. The small 'r' stands for 'root' and is a quirk of the DictEditor.
    first_part_schema = {
        "r": {
            "__collapse": False,
            "type": {
                "__html": "<h4>Property builder</h4>",
                "__description": "Choose the property you are interested in:",
                "__widget": "ToggleButtons",
                "__kwargs": {"options": ["P", "R", "S"]},
            },
            "__html": """<ul>
            <li><b>P (probability). </b> The probability that a path property (LTL, CTL etc.) holds.</li>
            <li><b>R (rewards). </b> The reward of the model under some constraint.</li>
            <li><b>S (steady-state behaviour). </b> The probability that a property holds, while the amount of steps tends to infinity.</li>
            </ul>""",
        },
    }

    # Choose an optimization direction. Relevant if the model has actions.
    opt = {
        "__description": "The model has actions, choose an optimization direction:",
        "__widget": "ToggleButtons",
        "__kwargs": {"options": ["max", "min"]},
    }

    # Explanation for operands.
    operand_explanation = 'An operand may use either labels, or variables with bounds. For example: "init" means that the state has label "init", and x>=5 means that the value of variable x is at least 5.'

    # Explanation for time bounds.
    time_bound_explanation = 'The time bound is used for bounded properties. For example "<=10" means that the property must hold within 10 time steps. Leave empty for unbounded properties.'

    # Choose a path property. Relevant if P was chosen.
    path_prop_schema = {
        "__html": "<h4>Path property</h4>",
        "r": {
            "__collapse": False,
            "operator": {
                "__description": "Choose the operator",
                "__widget": "ToggleButtons",
                "__kwargs": {"options": ["F", "G", "X", "U"]},
            },
            "__html": f"""<ul>
                <li><b>F (eventually). </b> The probability that operand 1 will eventually hold.</li>
                <li><b>G (always). </b> The probability that operand 1 always holds.</li>
                <li><b>X (next). </b> The probability that operand 1 holds in the next step.</li>
                <li><b>U (until). </b> The probability that operand 1 holds, until operand 2 holds.</li>
                </ul>
                {operand_explanation}""",
            "operand1": {
                "__description": "Operand 1",
                "__widget": "Text",
            },
            "operand2": {
                "__description": "Operand 2",
                "__widget": "Text",
            },
            "time_bound": {
                "__html": time_bound_explanation,
                "__description": "Time bound",
                "__widget": "Text",
            },
        },
    }

    # Choose a reward property. Relevant if R was chosen.
    reward_prop_schema = {
        "__html": "<h4>Reward property</h4>",
        "r": {
            "__collapse": False,
            "operator": {
                "__description": "Choose the operator",
                "__widget": "ToggleButtons",
                "__kwargs": {"options": ["F", "C", "LRA"]},
            },
            "__html": f"""<ul>
                <li><b>F (eventually). </b> The reward that is accumulated until the operand holds.</li>
                <li><b>C (total, cummulative reward). </b> The total reward that is accumulated when the system runs forever.</li>
                <li><b>LRA (gain / mean payoff / long run average). </b> The average expected reward per step if the system runs forever.</li>
                </ul>
                {operand_explanation}""",
            "operand1": {
                "__description": "Operand",
                "__widget": "Text",
            },
            "time_bound": {
                "__html": time_bound_explanation,
                "__description": "Time bound",
                "__widget": "Text",
            },
        },
    }

    # Select a reward model. Relevant if R was chosen, and there are multiple reward models.
    reward_model = {
        "__html": "The model has multiple reward models, select one",
        "__description": "Rew. model",
        "__widget": "Dropdown",
        "__kwargs": {"options": [""], "width": "auto"},
    }

    # Select a steady-state property. Relevant if S was chosen.
    steady_prop_schema = {
        "r": {
            "__collapse": False,
            "operand": {
                "__html": f"<h4>Steady-state property</h4>{operand_explanation}",
                "__description": "Operand",
                "__widget": "Text",
            },
        }
    }

    def values_to_path_reward_property(v: dict) -> str:
        """Convert the dictionary of values to a property string for path or reward properties."""
        op1 = v["operand1"].replace(" ", "")
        if v["operator"] == "U":
            op2 = v["operand2"].replace(" ", "")
            return f"[{op1} U{v['time_bound']} {op2}]"
        elif v["operator"] == "LRA":
            return f"[{v['operator']}]"
        elif v["operator"] == "C":
            return f"[{v['operator']}{v['time_bound']}]"
        else:
            return f"[{v['operator']}{v['time_bound']} {op1}]"

    def values_to_first_part_property(v: dict, reward_model="") -> str:
        """Convert the dictionary of values to a property string for the first part."""
        type_ = v["r"]["type"]
        opt_ = ""
        if model.supports_actions():
            opt_ = v["r"]["opt"]
        return (
            type_ + opt_ + "=?"
            if reward_model == ""
            else type_ + '{"' + reward_model + '"}' + opt_ + "=?"
        )

    if model.supports_actions():
        first_part_schema["r"]["opt"] = opt
    if len(model.rewards) > 1:
        old_operator = reward_prop_schema["r"]
        new_operator = {"reward_model": reward_model}
        new_operator["reward_model"]["__kwargs"]["options"] = [
            x.name for x in model.rewards
        ]
        new_operator.update(old_operator)  # Necessary to insert in front of the dict.

        reward_prop_schema["r"] = new_operator

    first_part_output = widgets.Output()
    second_part_output = widgets.Output()
    property_string_output = widgets.Output()

    class Values:
        # Small 'r' stands for 'root'.
        first_part_values = {
            "r": {
                "type": "P",
                "opt": "max",
            }
        }

        path_prop_values = {
            "r": {
                "operator": "F",
                "operand1": '"init"',
                "operand2": "x>=5",
                "time_bound": "",
            }
        }
        reward_prop_values = {
            "r": {
                "reward_model": "",
                "operator": "F",
                "operand1": '"init"',
                "time_bound": "",
            }
        }

        steady_prop_values = {"r": {"operand": "x>=5"}}

        def on_update_first_part(self):
            with second_part_output:
                ipd.clear_output()
            type_ = self.first_part_values["r"]["type"]
            if type_ == "P":
                de_path_prop.show()
            if type_ == "R":
                de_reward_prop.show()
            if type_ == "S":
                de_steady_prop.show()

        def on_update_path_prop(self):
            with property_string_output:
                ipd.clear_output()
                first_part = values_to_first_part_property(self.first_part_values)
                second_part = values_to_path_reward_property(self.path_prop_values["r"])
                print(first_part + " " + second_part)

        def on_update_reward_prop(self):
            with property_string_output:
                ipd.clear_output()
                first_part = values_to_first_part_property(self.first_part_values)
                second_part = values_to_path_reward_property(
                    self.reward_prop_values["r"]
                )
                print(first_part + " " + second_part)

        def on_update_steady_prop(self):
            with property_string_output:
                ipd.clear_output()
                first_part = values_to_first_part_property(self.first_part_values)
                second_part = "[" + self.steady_prop_values["r"]["operand"] + "]"
                print(first_part + " " + second_part)

    v = Values()

    de_first_part = DictEditor(
        first_part_schema,
        v.first_part_values,
        v.on_update_first_part,
        do_display=False,
        output=first_part_output,
    )
    de_path_prop = DictEditor(
        path_prop_schema,
        v.path_prop_values,
        v.on_update_path_prop,
        do_display=False,
        output=second_part_output,
    )
    de_reward_prop = DictEditor(
        reward_prop_schema,
        v.reward_prop_values,
        v.on_update_reward_prop,
        do_display=False,
        output=second_part_output,
    )
    de_steady_prop = DictEditor(
        steady_prop_schema,
        v.steady_prop_values,
        v.on_update_steady_prop,
        do_display=False,
        output=second_part_output,
    )

    de_first_part.show()

    ipd.display(first_part_output)
    ipd.display(second_part_output)
    ipd.display(property_string_output)



import io
import re

"""Used to autoscale an svg to remove unused space from the image."""


def remove_invalid_paths(svg_string: str) -> str:
    """Remove <path> elements that have no 'd' attribute"""
    return re.sub(r"<path(?![^>]* d=)[^>]*/>", "", svg_string)


def autoscale_svg(raw_svg: str, target_width: float) -> str:
    """Autoscale an SVG string to the target width while maintaining aspect ratio."""
    from lxml import etree
    from svgpathtools import svg2paths2

    # Parse the SVG from the raw string
    # Using lxml to parse the raw SVG string
    clean_svg = remove_invalid_paths(raw_svg)
    root = etree.fromstring(clean_svg)

    # Extract paths and calculate the bounding box using svgpathtools
    output_tuple = svg2paths2(io.StringIO(clean_svg))
    assert len(output_tuple) == 3
    paths, attributes, svg_attr = output_tuple

    # Calculate the bounding box of all paths
    xmin, xmax, ymin, ymax = 0, 0, 0, 0
    for i, path in enumerate(paths):
        atr = attributes[i]
        if atr["stroke"] != "none" and "d" in atr:
            box = path.bbox()
            x0, x1, y0, y1 = box
            xmin = x0 if xmin is None else min(xmin, x0)
            xmax = x1 if xmax is None else max(xmax, x1)
            ymin = y0 if ymin is None else min(ymin, y0)
            ymax = y1 if ymax is None else max(ymax, y1)
    width = xmax - xmin + 10  # Add small margin of 1 to avoid clipping
    height = ymax - ymin + 10

    # Set the viewBox and the width/height attributes to match the content size
    root.attrib["viewBox"] = f"{xmin} {ymin} {width} {height}"
    root.attrib["width"] = str(width)
    root.attrib["height"] = str(height)

    # Now scale according to the provided target width while keeping the aspect ratio
    aspect_ratio = height / width
    new_width = target_width
    new_height = target_width * aspect_ratio

    # Update width and height attributes in the SVG
    root.attrib["width"] = str(new_width)
    root.attrib["height"] = str(new_height)

    # Return the modified SVG as a string
    return etree.tostring(
        root,
        pretty_print=True,
        xml_declaration=True,
        encoding="utf-8",
    ).decode()
